# -*- coding: utf-8 -*-
# Generated by Django 1.11.2 on 2017-06-14 11:22
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='MyGraph',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=64)),
                ('url', models.TextField()),
            ],
        ),
        migrations.CreateModel(
            name='Profile',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('history', models.TextField(default=b'')),
                ('advancedUI', models.BooleanField(default=False)),
                ('user', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
            ],
        ),
        migrations.CreateModel(
            name='Variable',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=64)),
                ('value', models.CharField(max_length=64)),
                ('profile', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='account.Profile')),
            ],
        ),
        migrations.CreateModel(
            name='View',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=64)),
                ('profile', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='account.Profile')),
            ],
        ),
        migrations.CreateModel(
            name='Window',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=64)),
                ('top', models.IntegerField()),
                ('left', models.IntegerField()),
                ('width', models.IntegerField()),
                ('height', models.IntegerField()),
                ('url', models.TextField()),
                ('interval', models.IntegerField(null=True)),
                ('view', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='account.View')),
            ],
        ),
        migrations.AddField(
            model_name='mygraph',
            name='profile',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='account.Profile'),
        ),
    ]




from tagging.managers import (ModelTaggedItemManager as BaseModelTaggedItemManager,
                            TaggedItem)


class ContentTypeMixin(object):
    def with_intersection(self, tags, queryset=None):
        if queryset is None:
            return TaggedItem.objects.get_intersection_by_model(self.model, tags)
        else:
            return TaggedItem.objects.get_intersection_by_model(queryset, tags)


class ModelTaggedItemManager(ContentTypeMixin, BaseModelTaggedItemManager):
    pass



"""Copyright 2008 Orbitz WorldWide

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License."""

import os
import logging
from logging.handlers import TimedRotatingFileHandler as Rotater
from logging import NullHandler, FileHandler, StreamHandler

from django.conf import settings


class GraphiteLogger:
    def __init__(self):
        self.infoLogger = self._config_logger(
            settings.LOG_FILE_INFO,
            'info',
            settings.LOG_INFO_PERFORMANCE,
            level = logging.DEBUG if settings.DEBUG else logging.INFO,
        )
        self.exceptionLogger = self._config_logger(
            settings.LOG_FILE_EXCEPTION,
            'exception',
            True,
        )
        self.cacheLogger = self._config_logger(
            settings.LOG_FILE_CACHE,
            'cache',
            settings.LOG_CACHE_PERFORMANCE,
            level = logging.DEBUG if settings.DEBUG else logging.INFO,
        )
        self.renderingLogger = self._config_logger(
            settings.LOG_FILE_RENDERING,
            'rendering',
            settings.LOG_RENDERING_PERFORMANCE,
            level = logging.DEBUG if settings.DEBUG else logging.INFO,
        )

    @staticmethod
    def _config_logger(log_file_name, name, activate,
                       level=None, when='midnight',
                       backupCount=settings.LOG_ROTATION_COUNT):
        logger = logging.getLogger(name)
        if level is not None:
            logger.setLevel(level)
        if activate:  # if want to log this one
            if log_file_name == '-':
                formatter = logging.Formatter(
                    fmt='[%(asctime)s.%(msecs)03d] %(name)s %(levelname)s %(message)s',
                    datefmt='%d/%b/%Y %H:%M:%S')
                handler = StreamHandler()
            else:
                formatter = logging.Formatter(
                    fmt='%(asctime)s.%(msecs)03d :: %(message)s',
                    datefmt='%Y-%m-%d,%H:%M:%S')
                log_file = os.path.join(settings.LOG_DIR, log_file_name)
                if settings.LOG_ROTATION:  # if we want to rotate logs
                    handler = Rotater(log_file, when=when, backupCount=backupCount)
                else:  # let someone else, e.g. logrotate, rotate the logs
                    handler = FileHandler(log_file)
            handler.setFormatter(formatter)
            logger.addHandler(handler)
        else:
            logger.addHandler(NullHandler())
        return logger

    def info(self,msg,*args,**kwargs):
        return self.infoLogger.info(msg,*args,**kwargs)

    def debug(self,msg,*args,**kwargs):
        return self.infoLogger.debug(msg,*args,**kwargs)

    def warning(self,msg,*args,**kwargs):
        return self.infoLogger.warn(msg,*args,**kwargs)

    def exception(self,msg="Exception Caught",**kwargs):
        return self.exceptionLogger.exception(msg,**kwargs)

    def cache(self,msg,*args,**kwargs):
        return self.cacheLogger.info(msg,*args,**kwargs)

    def rendering(self,msg,*args,**kwargs):
        return self.renderingLogger.info(msg,*args,**kwargs)


log = GraphiteLogger() # import-shared logger instance



from threading import Lock

from graphite.render.grammar_unsafe import grammar as _grammar


class ThreadSafeGrammar(object):
    def __init__(self):
        self._lock = Lock()

    def parseString(self, instring):
        with self._lock:
            return _grammar.parseString(instring)


grammar = ThreadSafeGrammar()



import time
import socket
import struct
import random

from django.conf import settings

from graphite.render.hashing import ConsistentHashRing
from graphite.logger import log
from graphite.util import load_module, unpickle, parseHosts
from graphite.singleton import ThreadSafeSingleton


try:
    import six.moves.cPickle as pickle
except ImportError:
    import pickle


def load_keyfunc():
    if settings.CARBONLINK_HASHING_KEYFUNC:
        module_path, func_name = settings.CARBONLINK_HASHING_KEYFUNC.rsplit(':', 1)
        log.cache("Using keyfunc %s found in %s" % (str(func_name), str(module_path)))
        return load_module(module_path, member=func_name)
    else:
        return lambda x: x


class CarbonLinkRequestError(Exception):
    pass


class CarbonLinkPool(object):
    def __init__(self, hosts, timeout):
        self.hosts = [ (server, instance) for (server, port, instance) in hosts ]
        self.ports = { (server, instance): port for (server, port, instance) in hosts }
        self.timeout = float(timeout)
        servers = set([server for (server, port, instance) in hosts])
        if len(servers) < settings.REPLICATION_FACTOR:
            raise Exception("REPLICATION_FACTOR=%d cannot exceed servers=%d" % (
                settings.REPLICATION_FACTOR, len(servers)))

        self.hash_ring = ConsistentHashRing(
            self.hosts, hash_type=settings.CARBONLINK_HASHING_TYPE)
        self.keyfunc = load_keyfunc()
        self.connections = {}
        self.last_failure = {}
        # Create a connection pool for each host
        for host in self.hosts:
            self.connections[host] = set()

    def select_host(self, metric):
        "Returns the carbon host that has data for the given metric"
        key = self.keyfunc(metric)
        nodes = []
        servers = set()
        for node in self.hash_ring.get_nodes(key):
            (server, instance) = node
            if server in servers:
                continue
            servers.add(server)
            nodes.append(node)
            if len(servers) >= settings.REPLICATION_FACTOR:
                break

        available = [ n for n in nodes if self.is_available(n) ]
        return random.choice(available or nodes)

    def is_available(self, host):
        now = time.time()
        last_fail = self.last_failure.get(host, 0)
        return (now - last_fail) < settings.CARBONLINK_RETRY_DELAY

    def get_connection(self, host):
        # First try to take one out of the pool for this host
        (server, instance) = host
        port = self.ports[host]
        connectionPool = self.connections[host]
        try:
            return connectionPool.pop()
        except KeyError:
            pass #nothing left in the pool, gotta make a new connection

        log.cache("CarbonLink creating a new socket for %s" % str(host))
        try:
            connection = socket.create_connection((server, port), self.timeout)
        except socket.error:
            self.last_failure[host] = time.time()
            raise
        else:
            connection.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
            return connection

    def query(self, metric):
        request = dict(type='cache-query', metric=metric)
        results = self.send_request(request)
        log.cache("CarbonLink cache-query request for %s returned %d datapoints" % (
            metric, len(results['datapoints'])))
        return results['datapoints']

    def get_metadata(self, metric, key):
        request = dict(type='get-metadata', metric=metric, key=key)
        results = self.send_request(request)
        log.cache("CarbonLink get-metadata request received for %s:%s" % (metric, key))
        return results['value']

    def set_metadata(self, metric, key, value):
        request = dict(type='set-metadata', metric=metric, key=key, value=value)
        results = self.send_request(request)
        log.cache("CarbonLink set-metadata request received for %s:%s" % (metric, key))
        return results

    def send_request(self, request):
        metric = request['metric']
        serialized_request = pickle.dumps(request, protocol=settings.CARBONLINK_PICKLE_PROTOCOL)
        len_prefix = struct.pack("!L", len(serialized_request))
        request_packet = len_prefix + serialized_request
        result = {}
        result.setdefault('datapoints', [])

        if metric.startswith(settings.CARBON_METRIC_PREFIX):
            return self.send_request_to_all(request)

        if not self.hosts:
            log.cache("CarbonLink is not connected to any host. Returning empty nodes list")
            return result

        host = self.select_host(metric)
        conn = self.get_connection(host)
        log.cache("CarbonLink sending request for %s to %s" % (metric, str(host)))
        try:
            conn.sendall(request_packet)
            result = self.recv_response(conn)
        except Exception as e:
            self.last_failure[host] = time.time()
            log.cache("Exception getting data from cache %s: %s" % (str(host), e))
        else:
            self.connections[host].add(conn)
            if 'error' in result:
                log.cache("Error getting data from cache: %s" % result['error'])
                raise CarbonLinkRequestError(result['error'])
            log.cache("CarbonLink finished receiving %s from %s" % (str(metric), str(host)))
        return result

    def send_request_to_all(self, request):
        metric = request['metric']
        serialized_request = pickle.dumps(request, protocol=settings.CARBONLINK_PICKLE_PROTOCOL)
        len_prefix = struct.pack("!L", len(serialized_request))
        request_packet = len_prefix + serialized_request
        results = {}
        results.setdefault('datapoints', {})

        for host in self.hosts:
            conn = self.get_connection(host)
            log.cache("CarbonLink sending request for %s to %s" % (metric, str(host)))
            try:
                conn.sendall(request_packet)
                result = self.recv_response(conn)
            except Exception as e:
                self.last_failure[host] = time.time()
                log.cache("Exception getting data from cache %s: %s" % (str(host), e))
            else:
                self.connections[host].add(conn)
                if 'error' in result:
                    log.cache("Error getting data from cache %s: %s" % (str(host), result['error']))
                else:
                    if len(result['datapoints']) > 1:
                        results['datapoints'].update(result['datapoints'])
            log.cache("CarbonLink finished receiving %s from %s" % (str(metric), str(host)))

        return results

    def recv_response(self, conn):
        len_prefix = self.recv_exactly(conn, 4)
        body_size = struct.unpack("!L", len_prefix)[0]
        body = self.recv_exactly(conn, body_size)
        return unpickle.loads(body)

    @staticmethod
    def recv_exactly(conn, num_bytes):
        buf = b''
        while len(buf) < num_bytes:
            data = conn.recv(num_bytes - len(buf))
            if not data:
                raise Exception("Connection lost")
            buf += data

        return buf


@ThreadSafeSingleton
class GlobalCarbonLinkPool(CarbonLinkPool):
    def __init__(self):
        hosts = parseHosts(settings.CARBONLINK_HOSTS)
        timeout = settings.CARBONLINK_TIMEOUT
        CarbonLinkPool.__init__(self, hosts, timeout)


def CarbonLink():
    """Handy accessor for the global singleton."""
    return GlobalCarbonLinkPool.instance()



# -*- coding: utf-8 -*-
# Generated by Django 1.11.2 on 2017-06-14 11:22
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('account', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='Dashboard',
            fields=[
                ('name', models.CharField(max_length=128, primary_key=True, serialize=False)),
                ('state', models.TextField()),
                ('owners', models.ManyToManyField(related_name='dashboards', to='account.Profile')),
            ],
        ),
        migrations.CreateModel(
            name='Template',
            fields=[
                ('name', models.CharField(max_length=128, primary_key=True, serialize=False)),
                ('state', models.TextField()),
                ('owners', models.ManyToManyField(related_name='templates', to='account.Profile')),
            ],
        ),
    ]



from django.conf.urls import url
from graphite.functions.views import functionList, functionDetails

urlpatterns = [
  url(r'^/(.+)$', functionDetails, name='functionDetails'),
  url(r'^/?$', functionList, name='functionList'),
]



from __future__ import absolute_import
import os
import random
import sys
import time
import traceback
import types

from collections import defaultdict
from copy import deepcopy
from shutil import move
from tempfile import mkstemp

from django.conf import settings
from django.core.cache import cache
import six

try:
    from importlib import import_module
except ImportError:  # python < 2.7 compatibility
    from django.utils.importlib import import_module

from graphite.logger import log
from graphite.errors import InputParameterError
from graphite.node import LeafNode
from graphite.intervals import Interval, IntervalSet
from graphite.finders.utils import FindQuery, BaseFinder
from graphite.readers import MultiReader
from graphite.worker_pool.pool import get_pool, pool_exec, Job, PoolTimeoutError
from graphite.render.grammar import grammar


def get_finders(finder_path):
    module_name, class_name = finder_path.rsplit('.', 1)
    module = import_module(module_name)
    cls = getattr(module, class_name)

    if getattr(cls, 'factory', None):
        return cls.factory()

    # monkey patch so legacy finders will work
    finder = cls()
    if sys.version_info[0] >= 3:
        finder.fetch = types.MethodType(BaseFinder.fetch, finder)
        finder.find_multi = types.MethodType(BaseFinder.find_multi, finder)
        finder.get_index = types.MethodType(BaseFinder.get_index, finder)
    else:
        finder.fetch = types.MethodType(BaseFinder.fetch.__func__, finder)
        finder.find_multi = types.MethodType(BaseFinder.find_multi.__func__, finder)
        finder.get_index = types.MethodType(BaseFinder.get_index.__func__, finder)

    return [finder]


def get_tagdb(tagdb_path):
    module_name, class_name = tagdb_path.rsplit('.', 1)
    module = import_module(module_name)
    return getattr(module, class_name)(settings, cache=cache, log=log)


class Store(object):
    def __init__(self, finders=None, tagdb=None):
        if finders is None:
            finders = []
            for finder_path in settings.STORAGE_FINDERS:
                finders.extend(get_finders(finder_path))

        self.finders = finders

        if tagdb is None:
            tagdb = get_tagdb(settings.TAGDB or 'graphite.tags.base.DummyTagDB')
        self.tagdb = tagdb

    def get_finders(self, local=False):
        for finder in self.finders:
            # Support legacy finders by defaulting to 'disabled = False'
            if getattr(finder, 'disabled', False):
                continue

            # Support legacy finders by defaulting to 'local = True'
            if local and not getattr(finder, 'local', True):
                continue

            yield finder

    def pool_exec(self, jobs, timeout):
        if not jobs:
            return []

        thread_count = 0
        if settings.USE_WORKER_POOL:
            thread_count = min(len(self.finders), settings.POOL_MAX_WORKERS)

        return pool_exec(get_pool('finders', thread_count), jobs, timeout)

    def wait_jobs(self, jobs, timeout, context):
        if not jobs:
            return []

        start = time.time()
        results = []
        failed = []
        done = 0
        try:
            for job in self.pool_exec(jobs, timeout):
                elapsed = time.time() - start
                done += 1
                if job.exception:
                    failed.append(job)
                    log.info("Exception during %s after %fs: %s" % (
                        job, elapsed, str(job.exception))
                    )
                else:
                    log.debug("Got a result for %s after %fs" % (job, elapsed))
                    results.append(job.result)
        except PoolTimeoutError:
            message = "Timed out after %fs for %s" % (
                time.time() - start, context
            )
            log.info(message)
            if done == 0:
                raise Exception(message)

        if len(failed) == done:
            message = "All requests failed for %s (%d)" % (
                context, len(failed)
            )
            for job in failed:
                message += "\n\n%s: %s: %s" % (
                    job, job.exception,
                    '\n'.join(traceback.format_exception(*job.exception_info))
                )
            raise Exception(message)

        if len(results) < len(jobs) and settings.STORE_FAIL_ON_ERROR:
            message = "%s request(s) failed for %s (%d)" % (
                len(jobs) - len(results), context, len(jobs)
            )
            for job in failed:
                message += "\n\n%s: %s: %s" % (
                    job, job.exception,
                    '\n'.join(traceback.format_exception(*job.exception_info))
                )
            raise Exception(message)

        return results

    def fetch(self, patterns, startTime, endTime, now, requestContext):
        # deduplicate patterns
        patterns = sorted(set(patterns))

        if not patterns:
            return []

        log.debug(
            'graphite.storage.Store.fetch :: Starting fetch on all backends')

        jobs = []
        tag_patterns = None
        pattern_aliases = defaultdict(list)
        for finder in self.get_finders(requestContext.get('localOnly')):
            # if the finder supports tags, just pass the patterns through
            if getattr(finder, 'tags', False):
                job = Job(
                    finder.fetch, 'fetch for %s' % patterns,
                    patterns, startTime, endTime,
                    now=now, requestContext=requestContext
                )
                jobs.append(job)
                continue

            # if we haven't resolved the seriesByTag calls, build resolved patterns and translation table
            if tag_patterns is None:
                tag_patterns, pattern_aliases = self._tag_patterns(patterns, requestContext)

            # dispatch resolved patterns to finder
            job = Job(
                finder.fetch,
                'fetch for %s' % tag_patterns,
                tag_patterns, startTime, endTime,
                now=now, requestContext=requestContext
            )
            jobs.append(job)

        # Start fetches
        start = time.time()
        results = self.wait_jobs(jobs, settings.FETCH_TIMEOUT,
                                 'fetch for %s' % str(patterns))
        results = [i for l in results for i in l]  # flatten

        # translate path expressions for responses from resolved seriesByTag patterns
        for result in results:
            if result['name'] == result['pathExpression'] and result['pathExpression'] in pattern_aliases:
                for pathExpr in pattern_aliases[result['pathExpression']]:
                    newresult = deepcopy(result)
                    newresult['pathExpression'] = pathExpr
                    results.append(newresult)

        log.debug("Got all fetch results for %s in %fs" % (str(patterns), time.time() - start))
        return results

    def _tag_patterns(self, patterns, requestContext):
        tag_patterns = []
        pattern_aliases = defaultdict(list)

        for pattern in patterns:
            # if pattern isn't a seriesByTag call, just add it to the list
            if not pattern.startswith('seriesByTag('):
                tag_patterns.append(pattern)
                continue

            # perform the tagdb lookup
            exprs = tuple([
                t.string[1:-1]
                for t in grammar.parseString(pattern).expression.call.args
                if t.string
            ])
            taggedSeries = self.tagdb.find_series(exprs, requestContext=requestContext)
            if not taggedSeries:
                continue

            # add to translation table for path matching
            for series in taggedSeries:
                pattern_aliases[series].append(pattern)

            # add to list of resolved patterns
            tag_patterns.extend(taggedSeries)

        return sorted(set(tag_patterns)), pattern_aliases

    def get_index(self, requestContext=None):
        log.debug('graphite.storage.Store.get_index :: Starting get_index on all backends')

        if not requestContext:
            requestContext = {}

        context = 'get_index'
        jobs = [
            Job(finder.get_index, context, requestContext=requestContext)
            for finder in self.get_finders(local=requestContext.get('localOnly'))
        ]

        start = time.time()
        results = self.wait_jobs(jobs, settings.FETCH_TIMEOUT, context)
        results = [i for l in results if l is not None for i in l]  # flatten

        log.debug("Got all index results in %fs" % (time.time() - start))
        return sorted(list(set(results)))

    def find(self, pattern, startTime=None, endTime=None, local=False, headers=None, leaves_only=False):
        try:
            query = FindQuery(
                pattern, startTime, endTime,
                local=local,
                headers=headers,
                leaves_only=leaves_only
            )
        except Exception as e:
            raise InputParameterError(
                'Failed to instantiate find query: {err}'
                .format(err=str(e)))

        warn_threshold = settings.METRICS_FIND_WARNING_THRESHOLD
        fail_threshold = settings.METRICS_FIND_FAILURE_THRESHOLD

        matched_leafs = 0
        for match in self._find(query):
            if isinstance(match, LeafNode):
                matched_leafs += 1
            elif leaves_only:
                continue
            if matched_leafs > fail_threshold:
                raise Exception(
                    ("Query %s yields too many results and failed "
                     "(failure threshold is %d)") % (pattern, fail_threshold))
            yield match

        if matched_leafs > warn_threshold:
            log.warning(
                ("Query %s yields large number of results up to %d "
                 "(warning threshold is %d)") % (
                     pattern, matched_leafs, warn_threshold))

    def _find(self, query):
        context = 'find %s' % query
        jobs = [
            Job(finder.find_nodes, context, query)
            for finder in self.get_finders(query.local)
        ]

        # Group matching nodes by their path
        nodes_by_path = defaultdict(list)

        # Start finds
        start = time.time()
        results = self.wait_jobs(jobs, settings.FIND_TIMEOUT, context)
        for result in results:
            for node in result or []:
                nodes_by_path[node.path].append(node)

        log.debug("Got all find results for %s in %fs" % (
            str(query), time.time() - start)
        )
        return self._list_nodes(query, nodes_by_path)

    def _list_nodes(self, query, nodes_by_path):
        # Reduce matching nodes for each path to a minimal set
        found_branch_nodes = set()

        items = list(six.iteritems(nodes_by_path))
        random.shuffle(items)

        for path, nodes in items:
            leaf_nodes = []

            # First we dispense with the BranchNodes
            for node in nodes:
                if node.is_leaf:
                    leaf_nodes.append(node)
                # TODO need to filter branch nodes based on requested
                # interval... how?!?!?
                elif node.path not in found_branch_nodes:
                    yield node
                    found_branch_nodes.add(node.path)

            leaf_node = self._merge_leaf_nodes(query, path, leaf_nodes)
            if leaf_node:
                yield leaf_node

    def _merge_leaf_nodes(self, query, path, leaf_nodes):
        """Get a single node from a list of leaf nodes."""
        if not leaf_nodes:
            return None

        # Fast-path when there is a single node.
        if len(leaf_nodes) == 1:
            return leaf_nodes[0]

        # Calculate best minimal node set
        minimal_node_set = set()
        covered_intervals = IntervalSet([])

        # If the query doesn't fall entirely within the FIND_TOLERANCE window
        # we disregard the window. This prevents unnecessary remote fetches
        # caused when carbon's cache skews node.intervals, giving the appearance
        # remote systems have data we don't have locally, which we probably
        # do.
        now = int(time.time())
        tolerance_window = now - settings.FIND_TOLERANCE
        disregard_tolerance_window = query.interval.start < tolerance_window
        prior_to_window = Interval(float('-inf'), tolerance_window)

        def measure_of_added_coverage(
                node, drop_window=disregard_tolerance_window):
            relevant_intervals = node.intervals.intersect_interval(
                query.interval)
            if drop_window:
                relevant_intervals = relevant_intervals.intersect_interval(
                    prior_to_window)
            return covered_intervals.union(
                relevant_intervals).size - covered_intervals.size

        nodes_remaining = list(leaf_nodes)

        # Prefer local nodes first (and do *not* drop the tolerance window)
        for node in leaf_nodes:
            if node.local and measure_of_added_coverage(node, False) > 0:
                nodes_remaining.remove(node)
                minimal_node_set.add(node)
                covered_intervals = covered_intervals.union(node.intervals)

        if settings.REMOTE_STORE_MERGE_RESULTS:
            remote_nodes = [n for n in nodes_remaining if not n.local]
            for node in remote_nodes:
                nodes_remaining.remove(node)
                minimal_node_set.add(node)
                covered_intervals = covered_intervals.union(node.intervals)
        else:
            while nodes_remaining:
                node_coverages = [(measure_of_added_coverage(n), n)
                                  for n in nodes_remaining]
                best_coverage, best_node = max(node_coverages)

                if best_coverage == 0:
                    break

                nodes_remaining.remove(best_node)
                minimal_node_set.add(best_node)
                covered_intervals = covered_intervals.union(
                    best_node.intervals)

            # Sometimes the requested interval falls within the caching window.
            # We include the most likely node if the gap is within
            # tolerance.
            if not minimal_node_set:
                def distance_to_requested_interval(node):
                    if not node.intervals:
                        return float('inf')
                    latest = sorted(
                        node.intervals, key=lambda i: i.end)[-1]
                    distance = query.interval.start - latest.end
                    return distance if distance >= 0 else float('inf')

                best_candidate = min(
                    leaf_nodes, key=distance_to_requested_interval)
                if distance_to_requested_interval(
                        best_candidate) <= settings.FIND_TOLERANCE:
                    minimal_node_set.add(best_candidate)

        if not minimal_node_set:
            return None
        elif len(minimal_node_set) == 1:
            return minimal_node_set.pop()
        else:
            reader = MultiReader(minimal_node_set)
            return LeafNode(path, reader)

    def tagdb_auto_complete_tags(self, exprs, tagPrefix=None, limit=None, requestContext=None):
        log.debug(
            'graphite.storage.Store.auto_complete_tags :: Starting lookup on all backends')

        if requestContext is None:
            requestContext = {}

        context = 'tags for %s %s' % (str(exprs), tagPrefix or '')
        jobs = []
        use_tagdb = False
        for finder in self.get_finders(requestContext.get('localOnly')):
            if getattr(finder, 'tags', False):
                job = Job(
                    finder.auto_complete_tags, context,
                    exprs, tagPrefix=tagPrefix,
                    limit=limit, requestContext=requestContext
                )
                jobs.append(job)
            else:
                use_tagdb = True

        results = set()

        # if we're using the local tagdb then execute it (in the main thread
        # so that LocalDatabaseTagDB will work)
        if use_tagdb:
            results.update(self.tagdb.auto_complete_tags(
                exprs, tagPrefix=tagPrefix,
                limit=limit, requestContext=requestContext
            ))

        # Start fetches
        start = time.time()
        for result in self.wait_jobs(jobs, settings.FIND_TIMEOUT, context):
            results.update(result)

        # sort & limit results
        results = sorted(results)
        if limit:
            results = results[:int(limit)]

        log.debug("Got all autocomplete %s in %fs" % (
            context, time.time() - start)
        )
        return results

    def tagdb_auto_complete_values(self, exprs, tag, valuePrefix=None, limit=None, requestContext=None):
        log.debug(
            'graphite.storage.Store.auto_complete_values :: Starting lookup on all backends')

        if requestContext is None:
            requestContext = {}

        context = 'values for %s %s %s' % (str(exprs), tag, valuePrefix or '')
        jobs = []
        use_tagdb = False
        for finder in self.get_finders(requestContext.get('localOnly')):
            if getattr(finder, 'tags', False):
                job = Job(
                    finder.auto_complete_values, context,
                    exprs, tag, valuePrefix=valuePrefix,
                    limit=limit, requestContext=requestContext
                )
                jobs.append(job)
            else:
                use_tagdb = True

        # start finder jobs
        start = time.time()
        results = set()

        # if we're using the local tagdb then execute it (in the main thread
        # so that LocalDatabaseTagDB will work)
        if use_tagdb:
            results.update(self.tagdb.auto_complete_values(
                exprs, tag, valuePrefix=valuePrefix,
                limit=limit, requestContext=requestContext
            ))

        for result in self.wait_jobs(jobs, settings.FIND_TIMEOUT, context):
            results.update(result)

        # sort & limit results
        results = sorted(results)
        if limit:
            results = results[:int(limit)]

        log.debug("Got all autocomplete %s in %fs" % (
            context, time.time() - start)
        )
        return results


def extractForwardHeaders(request):
    headers = {}
    for name in settings.REMOTE_STORE_FORWARD_HEADERS:
        value = request.META.get('HTTP_%s' % name.upper().replace('-', '_'))
        if value is not None:
            headers[name] = value
    return headers


def write_index(index=None):
    if not index:
        index = settings.INDEX_FILE
    try:
        fd, tmp = mkstemp()
        try:
            tmp_index = os.fdopen(fd, 'wt')
            for metric in STORE.get_index():
                tmp_index.write("{0}\n".format(metric))
        finally:
            tmp_index.close()
        move(tmp, index)
    finally:
        try:
            os.unlink(tmp)
        except OSError:
            pass
    return None


STORE = Store()



from django.conf.urls import url
from . import views

urlpatterns = [
    url(r'^/save/(?P<name>[^/]+)$', views.save, name='dashboard_save'),
    url(r'^/save_template/(?P<name>[^/]+)/(?P<key>[^/]+)$', views.save_template,
        name='dashboard_save_template'),
    url(r'^/load/(?P<name>[^/]+)$', views.load, name='dashboard_load'),
    url(r'^/load/(?P<name>[^/]+)/(?P<val>[^/]+)$', views.load_template,
        name='dashboard_load_template'),
    url(r'^/load_template/(?P<name>[^/]+)/(?P<val>[^/]+)$', views.load_template,
        name='dashboard_load_template'),
    url(r'^/delete/(?P<name>[^/]+)$', views.delete, name='dashboard_delete'),
    url(r'^/create-temporary/?$', views.create_temporary, name='dashboard_create_temporary'),
    url(r'^/email$', views.email, name='dashboard_email'),
    url(r'^/find/?$', views.find, name='dashboard_find'),
    url(r'^/delete_template/(?P<name>[^/]+)$', views.delete_template,
        name='dashboard_delete_template'),
    url(r'^/find_template/?$', views.find_template, name='dashboard_find_template'),
    url(r'^/login/?$', views.user_login, name='dashboard_login'),
    url(r'^/logout/?$', views.user_logout, name='dashboard_logout'),
    url(r'^/help/?$', views.help, name='dashboard_help'),
    url(r'^/(?P<name>[^/]+)/(?P<val>[^/]+)$', views.template, name='dashboard_template'),
    url(r'^/(?P<name>[^/]+)$', views.dashboard, name='dashboard'),
    url(r'^/?$', views.dashboard, name='dashboard'),
]



"""Copyright 2008 Orbitz WorldWide

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License."""

import os
import pickle
from random import randint
from django.conf import settings

from graphite.compat import HttpResponse
from graphite.util import unpickle


def add(request):
    metrics = set( request.POST['metrics'].split() )
    whitelist = load_whitelist()
    new_whitelist = whitelist | metrics
    save_whitelist(new_whitelist)
    return HttpResponse(content_type="text/plain", content="OK")


def remove(request):
    metrics = set( request.POST['metrics'].split() )
    whitelist = load_whitelist()
    new_whitelist = whitelist - metrics
    save_whitelist(new_whitelist)
    return HttpResponse(content_type="text/plain", content="OK")


def show(request):
    whitelist = load_whitelist()
    members = '\n'.join( sorted(whitelist) )
    return HttpResponse(content_type="text/plain", content=members)


def load_whitelist():
    buffer = open(settings.WHITELIST_FILE, 'rb').read()
    whitelist = unpickle.loads(buffer)
    return whitelist


def save_whitelist(whitelist):
    # do this instead of dump() to raise potential exceptions before open()
    serialized = pickle.dumps(whitelist, protocol=-1)
    tmpfile = '%s-%d' % (settings.WHITELIST_FILE, randint(0, 100000))
    try:
        fh = open(tmpfile, 'wb')
        fh.write(serialized)
        fh.close()
        if os.path.exists(settings.WHITELIST_FILE):
            os.unlink(settings.WHITELIST_FILE)
        os.rename(tmpfile, settings.WHITELIST_FILE)
    finally:
        if os.path.exists(tmpfile):
            os.unlink(tmpfile)



"""Utility functions for tag databases."""
import re

from hashlib import sha256


class TaggedSeries(object):
    prohibitedTagChars = ';!^='

    @classmethod
    def validateTagAndValue(cls, tag, value):
        """validate the given tag / value based on the specs in the documentation"""
        if len(tag) == 0:
            raise Exception('Tag may not be empty')
        if len(value) == 0:
            raise Exception('Value for tag "{tag}" may not be empty'.format(tag=tag))

        for char in cls.prohibitedTagChars:
            if char in tag:
                raise Exception('Character "{}" is not allowed in tag "{}"'.format(char, tag))

        if ';' in value:
            raise Exception('Character ";" is not allowed in value "{}" of tag {}'.format(value, tag))

        if value[0] == '~':
            raise Exception('Tag values are not allowed to start with "~" in tag "{tag}"'.format(tag=tag))

    @classmethod
    def parse(cls, path):
        # if path is in openmetrics format: metric{tag="value",...}
        if path[-2:] == '"}' and '{' in path:
            return cls.parse_openmetrics(path)

        # path is a carbon path with optional tags: metric;tag=value;...
        return cls.parse_carbon(path)

    @classmethod
    def parse_openmetrics(cls, path):
        """parse a path in openmetrics format: metric{tag="value",...}

        https://github.com/RichiH/OpenMetrics
        """
        (metric, rawtags) = path[0:-1].split('{', 2)
        if not metric:
            raise Exception('Cannot parse path %s, no metric found' % path)

        tags = {}

        while len(rawtags) > 0:
            m = re.match(r'([^=]+)="((?:[\\]["\\]|[^"\\])+)"(:?,|$)', rawtags)
            if not m:
                raise Exception('Cannot parse path %s, invalid segment %s' % (path, rawtags))

            tag = m.group(1)
            value = m.group(2).replace(r'\"', '"').replace(r'\\', '\\')

            cls.validateTagAndValue(tag, value)

            tags[tag] = value
            rawtags = rawtags[len(m.group(0)):]

        tags['name'] = cls.sanitize_name_as_tag_value(metric)
        return cls(metric, tags)

    @classmethod
    def parse_carbon(cls, path):
        """parse a carbon path with optional tags: metric;tag=value;..."""
        segments = path.split(';')

        metric = segments[0]
        if not metric:
            raise Exception('Cannot parse path %s, no metric found' % path)

        tags = {}

        for segment in segments[1:]:
            tag = segment.split('=', 1)
            if len(tag) != 2 or not tag[0]:
                raise Exception('Cannot parse path %s, invalid segment %s' % (path, segment))

            cls.validateTagAndValue(*tag)

            tags[tag[0]] = tag[1]

        tags['name'] = cls.sanitize_name_as_tag_value(metric)
        return cls(metric, tags)

    @staticmethod
    def sanitize_name_as_tag_value(name):
        """take a metric name and sanitize it so it is guaranteed to be a valid tag value"""
        sanitized = name.lstrip('~')

        if len(sanitized) == 0:
            raise Exception('Cannot use metric name %s as tag value, results in emptry string' % (name))

        return sanitized

    @staticmethod
    def format(tags):
        return tags.get('name', '') + ''.join(sorted([
            ';%s=%s' % (tag, value)
            for tag, value in tags.items()
            if tag != 'name'
        ]))

    @staticmethod
    def encode(metric, sep='.', hash_only=False):
        """
        Helper function to encode tagged series for storage in whisper etc

        When tagged series are detected, they are stored in a separate hierarchy of folders under a
        top-level _tagged folder, where subfolders are created by using the first 3 hex digits of the
        sha256 hash of the tagged metric path (4096 possible folders), and second-level subfolders are
        based on the following 3 hex digits (another 4096 possible folders) for a total of 4096^2
        possible subfolders. The metric files themselves are created with any . in the metric path
        replaced with -, to avoid any issues where metrics, tags or values containing a '.' would end
        up creating further subfolders. This helper is used by both whisper and ceres, but by design
        each carbon database and graphite-web finder is responsible for handling its own encoding so
        that different backends can create their own schemes if desired.

        The hash_only parameter can be set to True to use the hash as the filename instead of a
        human-readable name.  This avoids issues with filename length restrictions, at the expense of
        being unable to decode the filename and determine the original metric name.

        A concrete example:

        .. code-block:: none

          some.metric;tag1=value2;tag2=value.2

          with sha256 hash starting effaae would be stored in:

          _tagged/eff/aae/some-metric;tag1=value2;tag2=value-2.wsp (whisper)
          _tagged/eff/aae/some-metric;tag1=value2;tag2=value-2 (ceres)

        """
        if ';' in metric:
            metric_hash = sha256(metric.encode('utf8')).hexdigest()
            return sep.join([
                '_tagged',
                metric_hash[0:3],
                metric_hash[3:6],
                metric_hash if hash_only else metric.replace('.', '_DOT_')
            ])

        # metric isn't tagged, just replace dots with the separator and trim any leading separator
        return metric.replace('.', sep).lstrip(sep)

    @staticmethod
    def decode(path, sep='.'):
        """
        Helper function to decode tagged series from storage in whisper etc
        """
        if path.startswith('_tagged'):
            return path.split(sep, 3)[-1].replace('_DOT_', '.')

        # metric isn't tagged, just replace the separator with dots
        return path.replace(sep, '.')

    def __init__(self, metric, tags, series_id=None):
        self.metric = metric
        self.tags = tags
        self.id = series_id

    @property
    def path(self):
        return self.__class__.format(self.tags)



"""Copyright 2008 Orbitz WorldWide

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License."""
# Django settings for graphite project.
# DO NOT MODIFY THIS FILE DIRECTLY - use local_settings.py instead
from __future__ import print_function
import os
import sys
from os.path import abspath, dirname, join
from warnings import warn
from importlib import import_module

from django import VERSION as DJANGO_VERSION
try:
    from django.urls import reverse_lazy
except ImportError:  # Django < 1.10
    from django.core.urlresolvers import reverse_lazy


GRAPHITE_WEB_APP_SETTINGS_LOADED = False
WEBAPP_VERSION = '1.1.10'
DEBUG = False
JAVASCRIPT_DEBUG = False

DATE_FORMAT = '%m/%d'

# Allow UTF-8 metrics' names (can cause performance issues)
UTF8_METRICS = False

# Filesystem layout
WEB_DIR = dirname( abspath(__file__) )
WEBAPP_DIR = dirname(WEB_DIR)
GRAPHITE_ROOT = dirname(WEBAPP_DIR)
# Initialize additional path variables
# Defaults for these are set after local_settings is imported
STATIC_ROOT = ''
STATIC_URL = '/static/'
URL_PREFIX = ''
CONF_DIR = ''
DASHBOARD_CONF = ''
GRAPHTEMPLATES_CONF = ''
STORAGE_DIR = ''
WHITELIST_FILE = ''
INDEX_FILE = ''
LOG_DIR = ''
CERES_DIR = ''
WHISPER_DIR = ''
RRD_DIR = ''
STANDARD_DIRS = []

# Timeout settings
FIND_TIMEOUT = None  # default 3.0 see below
FETCH_TIMEOUT = None  # default 6.0 see below

# Cluster settings
CLUSTER_SERVERS = []

# Worker Pool
USE_WORKER_POOL = True
POOL_MAX_WORKERS = 10

# This settings control whether https is used to communicate between cluster members
INTRACLUSTER_HTTPS = False
REMOTE_FIND_TIMEOUT = None  # Replaced by FIND_TIMEOUT
REMOTE_FETCH_TIMEOUT = None  # Replaced by FETCH_TIMEOUT
REMOTE_RETRY_DELAY = 60.0
REMOTE_EXCLUDE_LOCAL = False
STORE_FAIL_ON_ERROR = False
REMOTE_STORE_MERGE_RESULTS = True
REMOTE_STORE_FORWARD_HEADERS = []
REMOTE_STORE_USE_POST = False
REMOTE_BUFFER_SIZE = 1024 * 1024 # Set to 0 to prevent streaming deserialization

# Carbonlink settings
CARBON_METRIC_PREFIX='carbon'
CARBONLINK_HOSTS = ["127.0.0.1:7002"]
CARBONLINK_TIMEOUT = 1.0
CARBONLINK_HASHING_KEYFUNC = None
CARBONLINK_HASHING_TYPE = 'carbon_ch'
CARBONLINK_RETRY_DELAY = 15
CARBONLINK_PICKLE_PROTOCOL = -1
REPLICATION_FACTOR = 1

# Cache settings.
MEMCACHE_HOSTS = []
MEMCACHE_KEY_PREFIX = ''
MEMCACHE_OPTIONS = {}
CACHES={}
FIND_CACHE_DURATION = 300
FIND_TOLERANCE = 2 * FIND_CACHE_DURATION
DEFAULT_CACHE_DURATION = 60 #metric data and graphs are cached for one minute by default
DEFAULT_CACHE_POLICY = []

# this setting controls the default xFilesFactor used for query-time aggregration
DEFAULT_XFILES_FACTOR = 0

# These can also be configured using:
# https://docs.djangoproject.com/en/1.11/topics/logging/
LOG_RENDERING_PERFORMANCE = False
LOG_CACHE_PERFORMANCE = False
LOG_INFO_PERFORMANCE = True
LOG_ROTATION = True
LOG_ROTATION_COUNT = 1

LOG_FILE_INFO = 'info.log'
LOG_FILE_EXCEPTION = 'exception.log'
LOG_FILE_CACHE = 'cache.log'
LOG_FILE_RENDERING = 'rendering.log'

MAX_FETCH_RETRIES = 2

# This settings limit metrics find to prevent from too large query
METRICS_FIND_WARNING_THRESHOLD = float('Inf') # Print a warning if more than X metrics are returned
METRICS_FIND_FAILURE_THRESHOLD = float('Inf') # Fail if more than X metrics are returned

#Local rendering settings
RENDER_TRIM_RECENT_IN_AGGREGATE = False #if True, set most recent datapoints to None if prior datapoint was not None

#Remote rendering settings
REMOTE_RENDERING = False #if True, rendering is delegated to RENDERING_HOSTS
RENDERING_HOSTS = []
REMOTE_RENDER_CONNECT_TIMEOUT = 1.0

# Function settings
# automatically removing gaps from arguments of derivative functions
AUTOCOMPRESS_GAPS_IN_DERIVATIVE_FUNCTIONS = False #experimental, do not enable

#Miscellaneous settings
DOCUMENTATION_VERSION = 'latest' if 'dev' in WEBAPP_VERSION else WEBAPP_VERSION
DOCUMENTATION_URL = 'https://graphite.readthedocs.io/en/{}/'.format(DOCUMENTATION_VERSION)
ALLOW_ANONYMOUS_CLI = True
LEGEND_MAX_ITEMS = 10
RRD_CF = 'AVERAGE'
STORAGE_FINDERS = (
    'graphite.finders.remote.RemoteFinder',
    'graphite.finders.standard.StandardFinder',
)

# TagDB settings
TAGDB = 'graphite.tags.localdatabase.LocalDatabaseTagDB'

TAGDB_CACHE_DURATION = 60

TAGDB_AUTOCOMPLETE_LIMIT = 100

TAGDB_REDIS_HOST = 'localhost'
TAGDB_REDIS_PORT = 6379
TAGDB_REDIS_DB = 0
TAGDB_REDIS_PASSWORD = ''

TAGDB_HTTP_URL = ''
TAGDB_HTTP_USER = ''
TAGDB_HTTP_PASSWORD = ''
TAGDB_HTTP_AUTOCOMPLETE = False

# Function plugins
FUNCTION_PLUGINS = []


MIDDLEWARE = ()
if DJANGO_VERSION < (1, 10):
    MIDDLEWARE_CLASSES = MIDDLEWARE
MAX_TAG_LENGTH = 50
AUTO_REFRESH_INTERVAL = 60

#Authentication settings
USE_LDAP_AUTH = False
LDAP_SERVER = "" # "ldapserver.mydomain.com"
LDAP_PORT = 389
LDAP_USE_TLS = False
LDAP_SEARCH_BASE = "" # "OU=users,DC=mydomain,DC=com"
LDAP_BASE_USER = "" # "CN=some_readonly_account,DC=mydomain,DC=com"
LDAP_BASE_PASS = "" # "my_password"
LDAP_USER_QUERY = "" # "(username=%s)"  For Active Directory use "(sAMAccountName=%s)"
LDAP_URI = None
LDAP_USER_DN_TEMPLATE = None

#Set this to True to delegate authentication to the web server
USE_REMOTE_USER_AUTHENTICATION = False
REMOTE_USER_BACKEND = "" # Provide an alternate or subclassed backend
REMOTE_USER_MIDDLEWARE = "" # Provide an alternate or subclassed middleware
AUTHENTICATION_BACKENDS=[]

# Django 1.5 requires this so we set a default but warn the user
SECRET_KEY = 'UNSAFE_DEFAULT'

# Input validation
# - When False we still validate the received input parameters, but if validation
#   detects an issue it only logs an error and doesn't directly reject the request
# - When True we reject requests of which the input validation detected an issue with the
#   provided arguments and return an error message to the user
ENFORCE_INPUT_VALIDATION = False

# headers which shall be added to log statements informing about invalid queries,
# this is useful to identify where a query came from.
# The dict is keyed by the header name and the associated value is a short description
# of the header which will be used in the log statement, for example:
# {'X-FORWARD-FOR': 'forwarded-for'}
INPUT_VALIDATION_SOURCE_ID_HEADERS = {}

# Django 1.5 requires this to be set. Here we default to prior behavior and allow all
ALLOWED_HOSTS = [ '*' ]

# Override to link a different URL for login (e.g. for django_openid_auth)
LOGIN_URL = reverse_lazy('account_login')

# Set the default timezone to UTC
TIME_ZONE = 'UTC'

# Set to True to require authentication to save or delete dashboards
DASHBOARD_REQUIRE_AUTHENTICATION = False
# Require Django change/delete permissions to save or delete dashboards.
# NOTE: Requires DASHBOARD_REQUIRE_AUTHENTICATION to be set
DASHBOARD_REQUIRE_PERMISSIONS = False
# Name of a group to which the user must belong to save or delete dashboards.  Alternative to
# DASHBOARD_REQUIRE_PERMISSIONS, particularly useful when using only LDAP (without Admin app)
# NOTE: Requires DASHBOARD_REQUIRE_AUTHENTICATION to be set
DASHBOARD_REQUIRE_EDIT_GROUP = None

DATABASES = None

# If using rrdcached, set to the address or socket of the daemon
FLUSHRRDCACHED = ''

## Load our local_settings
SETTINGS_MODULE = os.environ.get('GRAPHITE_SETTINGS_MODULE', 'graphite.local_settings')
try:
    globals().update(import_module(SETTINGS_MODULE).__dict__)
except ImportError:
    print("Could not import {0}, using defaults!".format(SETTINGS_MODULE), file=sys.stderr)

## Load Django settings if they werent picked up in local_settings
if not GRAPHITE_WEB_APP_SETTINGS_LOADED:
    from graphite.app_settings import *  # noqa


STATICFILES_DIRS = (
    join(WEBAPP_DIR, 'content'),
)

# Handle renamed timeout settings
FIND_TIMEOUT = FIND_TIMEOUT or REMOTE_FIND_TIMEOUT or 3.0
FETCH_TIMEOUT = FETCH_TIMEOUT or REMOTE_FETCH_TIMEOUT or 6.0

## Set config dependent on flags set in local_settings
# Path configuration
if not STATIC_ROOT:
    STATIC_ROOT = join(GRAPHITE_ROOT, 'static')

if not CONF_DIR:
    CONF_DIR = os.environ.get('GRAPHITE_CONF_DIR', join(GRAPHITE_ROOT, 'conf'))
if not DASHBOARD_CONF:
    DASHBOARD_CONF = join(CONF_DIR, 'dashboard.conf')
if not GRAPHTEMPLATES_CONF:
    GRAPHTEMPLATES_CONF = join(CONF_DIR, 'graphTemplates.conf')

if not STORAGE_DIR:
    STORAGE_DIR = os.environ.get('GRAPHITE_STORAGE_DIR', join(GRAPHITE_ROOT, 'storage'))
if not WHITELIST_FILE:
    WHITELIST_FILE = join(STORAGE_DIR, 'lists', 'whitelist')
if not INDEX_FILE:
    INDEX_FILE = join(STORAGE_DIR, 'index')
if not LOG_DIR:
    LOG_DIR = join(STORAGE_DIR, 'log', 'webapp')
if not WHISPER_DIR:
    WHISPER_DIR = join(STORAGE_DIR, 'whisper/')
if not CERES_DIR:
    CERES_DIR = join(STORAGE_DIR, 'ceres/')
if not RRD_DIR:
    RRD_DIR = join(STORAGE_DIR, 'rrd/')
if not STANDARD_DIRS:
    try:
        import whisper  # noqa
        if os.path.exists(WHISPER_DIR):
            STANDARD_DIRS.append(WHISPER_DIR)
    except ImportError:
        print("WARNING: whisper module could not be loaded, whisper support disabled", file=sys.stderr)
    try:
        import ceres  # noqa
        if os.path.exists(CERES_DIR):
            STANDARD_DIRS.append(CERES_DIR)
    except ImportError:
        pass
    try:
        import rrdtool  # noqa
        if os.path.exists(RRD_DIR):
            STANDARD_DIRS.append(RRD_DIR)
    except ImportError:
        pass

if DATABASES is None:
    DATABASES = {
      'default': {
        'NAME': join(STORAGE_DIR, 'graphite.db'),
        'ENGINE': 'django.db.backends.sqlite3',
        'USER': '',
        'PASSWORD': '',
        'HOST': '',
        'PORT': '',
      },
    }

# Handle URL prefix in static files handling
if URL_PREFIX and not STATIC_URL.startswith(URL_PREFIX):
    STATIC_URL = '/{0}{1}'.format(URL_PREFIX.strip('/'), STATIC_URL)

# Default sqlite db file
# This is set here so that a user-set STORAGE_DIR is available
if 'sqlite3' in DATABASES.get('default',{}).get('ENGINE','') \
        and not DATABASES.get('default',{}).get('NAME'):
    DATABASES['default']['NAME'] = join(STORAGE_DIR, 'graphite.db')

# Caching shortcuts
if MEMCACHE_HOSTS:
    CACHES['default'] = {
        'BACKEND': 'django.core.cache.backends.memcached.MemcachedCache',
        'LOCATION': MEMCACHE_HOSTS,
        'TIMEOUT': DEFAULT_CACHE_DURATION,
        'KEY_PREFIX': MEMCACHE_KEY_PREFIX,
        'OPTIONS': MEMCACHE_OPTIONS,
    }

if not CACHES:
    CACHES = {
        'default': {
            'BACKEND': 'django.core.cache.backends.dummy.DummyCache',
        },
    }

# Authentication shortcuts
if USE_LDAP_AUTH and LDAP_URI is None:
    LDAP_URI = "ldap://%s:%d/" % (LDAP_SERVER, LDAP_PORT)

if USE_REMOTE_USER_AUTHENTICATION or REMOTE_USER_BACKEND:
    if REMOTE_USER_MIDDLEWARE:
        MIDDLEWARE += (REMOTE_USER_MIDDLEWARE,)
    else:
        MIDDLEWARE += ('django.contrib.auth.middleware.RemoteUserMiddleware',)
    if DJANGO_VERSION < (1, 10):
        MIDDLEWARE_CLASSES = MIDDLEWARE
    if REMOTE_USER_BACKEND:
        AUTHENTICATION_BACKENDS.insert(0,REMOTE_USER_BACKEND)
    else:
        AUTHENTICATION_BACKENDS.insert(0,'django.contrib.auth.backends.RemoteUserBackend')

if USE_LDAP_AUTH:
    AUTHENTICATION_BACKENDS.insert(0,'graphite.account.ldapBackend.LDAPBackend')

if SECRET_KEY == 'UNSAFE_DEFAULT':
    warn('SECRET_KEY is set to an unsafe default. This should be set in local_settings.py for better security')

USE_TZ = True



import re

from django.db import connection
from hashlib import sha256

from graphite.tags.base import BaseTagDB, TaggedSeries


class LocalDatabaseTagDB(BaseTagDB):
    def find_series_query(self, tags):
        # sql will select series that match all tag expressions that don't match empty tags
        sql = 'SELECT s.path'
        sql += ' FROM tags_series AS s'
        params = []

        where = []
        whereparams = []

        all_match_empty = True

        # expressions that do match empty tags will be used to filter the result
        filters = []

        i = 0
        for tagspec in tags:
            (tag, operator, spec) = self.parse_tagspec(tagspec)

            i += 1
            s = str(i)

            if operator == '=':
                matches_empty = spec == ''

                if not matches_empty:
                    where.append('v' + s + '.value=%s')
                    whereparams.append(spec)

            elif operator == '=~':
                # make sure regex is anchored
                if not spec.startswith('^'):
                    spec = '^(' + spec + ')'

                matches_empty = bool(re.match(spec, ''))

                if not matches_empty:
                    where.append('v' + s + '.value ' + self._regexp_operator(connection) + ' %s')
                    whereparams.append(spec)

            elif operator == '!=':
                matches_empty = spec != ''

                if not matches_empty:
                    where.append('v' + s + '.value<>%s')
                    whereparams.append(spec)

            elif operator == '!=~':
                # make sure regex is anchored
                if not spec.startswith('^'):
                    spec = '^(' + spec + ')'

                matches_empty = not re.match(spec, '')

                if not matches_empty:
                    where.append('v' + s + '.value ' + self._regexp_not_operator(connection) + ' %s')
                    whereparams.append(spec)

            else:
                raise ValueError("Invalid operator %s" % operator)

            if matches_empty:
                filters.append((tag, operator, spec))
            else:
                sql += ' JOIN tags_tag AS t' + s + ' ON t' + s + '.tag=%s'
                params.append(tag)
                sql += ' JOIN tags_seriestag AS st' + s + ' ON st' + s + '.series_id=s.id AND st' + s + '.tag_id=t' + s + '.id'
                sql += ' JOIN tags_tagvalue AS v' + s + ' ON v' + s + '.id=st' + s + '.value_id'

            all_match_empty = all_match_empty and matches_empty

        if all_match_empty:
            raise ValueError("At least one tagspec must not match the empty string")

        if where:
            sql += ' WHERE ' + ' AND '.join(where)
            params.extend(whereparams)

        sql += ' ORDER BY s.path'

        return sql, params, filters

    def _find_series(self, tags, requestContext=None):
        sql, params, filters = self.find_series_query(tags)

        def matches_filters(path):
            if not filters:
                return True

            try:
                parsed = self.parse(path)
            except Exception:
                return False

            for (tag, operator, spec) in filters:
                value = parsed.tags.get(tag, '')
                if (
                    (operator == '=' and value != spec) or
                    (operator == '=~' and re.match(spec, value) is None) or
                    (operator == '!=' and value == spec) or
                    (operator == '!=~' and re.match(spec, value) is not None)
                ):
                    return False

            return True

        with connection.cursor() as cursor:
            cursor.execute(sql, params)

            return [row[0] for row in cursor if matches_filters(row[0])]

    @staticmethod
    def _path_hash(path):
        return sha256(path.encode('utf8')).hexdigest()

    def get_series(self, path, requestContext=None):
        return self._get_series(self._path_hash(path))

    @staticmethod
    def _get_series(path_hash):
        with connection.cursor() as cursor:
            sql = 'SELECT s.id, t.tag, v.value'
            sql += ' FROM tags_series AS s'
            sql += ' JOIN tags_seriestag AS st ON st.series_id=s.id'
            sql += ' JOIN tags_tag AS t ON t.id=st.tag_id'
            sql += ' JOIN tags_tagvalue AS v ON v.id=st.value_id'
            sql += ' WHERE s.hash=%s'
            params = [path_hash]
            cursor.execute(sql, params)

            series_id = None

            tags = {tag: value for (series_id, tag, value) in cursor}

            if not tags:
                return None

            return TaggedSeries(tags['name'], tags, series_id=series_id)

    def list_tags(self, tagFilter=None, limit=None, requestContext=None):
        with connection.cursor() as cursor:
            sql = 'SELECT t.id, t.tag'
            sql += ' FROM tags_tag AS t'
            params = []

            if tagFilter:
                # make sure regex is anchored
                if not tagFilter.startswith('^'):
                    tagFilter = '^(' + tagFilter + ')'
                sql += ' WHERE t.tag ' + self._regexp_operator(connection) + ' %s'
                params.append(tagFilter)

            sql += ' ORDER BY t.tag'

            if limit:
                sql += ' LIMIT %s'
                params.append(int(limit))

            cursor.execute(sql, params)

            return [{'id': tag_id, 'tag': tag} for (tag_id, tag) in cursor]

    def get_tag(self, tag, valueFilter=None, limit=None, requestContext=None):
        with connection.cursor() as cursor:
            sql = 'SELECT t.id, t.tag'
            sql += ' FROM tags_tag AS t'
            sql += ' WHERE t.tag=%s'
            params = [tag]
            cursor.execute(sql, params)

            row = cursor.fetchone()

        if not row:
            return None

        (tag_id, tag) = row

        return {
            'id': tag_id,
            'tag': tag,
            'values': self.list_values(
                tag,
                valueFilter=valueFilter,
                limit=limit,
                requestContext=requestContext
            ),
        }

    def list_values(self, tag, valueFilter=None, limit=None, requestContext=None):
        with connection.cursor() as cursor:
            sql = 'SELECT v.id, v.value, COUNT(st.id)'
            sql += ' FROM tags_tagvalue AS v'
            sql += ' JOIN tags_seriestag AS st ON st.value_id=v.id'
            sql += ' JOIN tags_tag AS t ON t.id=st.tag_id'
            sql += ' WHERE t.tag=%s'
            params = [tag]

            if valueFilter:
                # make sure regex is anchored
                if not valueFilter.startswith('^'):
                    valueFilter = '^(' + valueFilter + ')'
                sql += ' AND v.value ' + self._regexp_operator(connection) + ' %s'
                params.append(valueFilter)

            sql += ' GROUP BY v.id, v.value'
            sql += ' ORDER BY v.value'

            if limit:
                sql += ' LIMIT %s'
                params.append(int(limit))

            cursor.execute(sql, params)

            return [{'id': value_id, 'value': value, 'count': count} for (value_id, value, count) in cursor]

    @staticmethod
    def _insert_ignore(table, cols, data):
        sql = table + ' (' + ','.join(cols) + ') VALUES ' + ', '.join(['(' + ', '.join(['%s'] * len(cols)) + ')'] * len(data))
        params = []
        for row in data:
            params.extend(row)

        if connection.vendor == 'mysql':
            sql = 'INSERT IGNORE INTO ' + sql
        elif connection.vendor == 'sqlite':
            sql = 'INSERT OR IGNORE INTO ' + sql
        elif connection.vendor == 'postgresql':
            sql = 'INSERT INTO ' + sql + ' ON CONFLICT DO NOTHING'  # nosec
        else:
            raise Exception('Unsupported database vendor ' + connection.vendor)

        with connection.cursor() as cursor:
            cursor.execute(sql, params)

    @staticmethod
    def _regexp_operator(connection):
        if connection.vendor == 'mysql':
            return 'REGEXP'
        if connection.vendor == 'sqlite':
            # django provides an implementation of REGEXP for sqlite
            return 'REGEXP'
        if connection.vendor == 'postgresql':
            return '~*'
        raise Exception('Database vendor ' + connection.vendor + ' does not support regular expressions')

    @staticmethod
    def _regexp_not_operator(connection):
        if connection.vendor == 'mysql':
            return 'NOT REGEXP'
        if connection.vendor == 'sqlite':
            # django provides an implementation of REGEXP for sqlite
            return 'NOT REGEXP'
        if connection.vendor == 'postgresql':
            return '!~*'
        raise Exception('Database vendor ' + connection.vendor + ' does not support regular expressions')

    def tag_series(self, series, requestContext=None):
        # extract tags and normalize path
        parsed = self.parse(series)

        path = parsed.path
        path_hash = self._path_hash(path)

        # check if path is already tagged
        curr = self._get_series(path_hash)
        if curr and parsed.tags == curr.tags:
            return path

        with connection.cursor() as cursor:
            # tags
            self._insert_ignore('tags_tag', ['tag'], [[tag] for tag in parsed.tags.keys()])

            sql = 'SELECT id, tag FROM tags_tag WHERE tag IN (' + ', '.join(['%s'] * len(parsed.tags)) + ')'  # nosec
            params = list(parsed.tags.keys())
            cursor.execute(sql, params)
            tag_ids = {tag: tag_id for (tag_id, tag) in cursor}

            # tag values
            self._insert_ignore('tags_tagvalue', ['value'], [[value] for value in parsed.tags.values()])

            sql = 'SELECT id, value FROM tags_tagvalue WHERE value IN (' + ', '.join(['%s'] * len(parsed.tags)) + ')'  # nosec
            params = list(parsed.tags.values())
            cursor.execute(sql, params)
            value_ids = {value: value_id for (value_id, value) in cursor}

            # series
            if curr:
                series_id = curr.id
            else:
                # hash column is used to support a unique index in mysql since path can be longer than 191 characters
                self._insert_ignore('tags_series', ['hash', 'path'], [[path_hash, path]])

                sql = 'SELECT id FROM tags_series WHERE hash=%s'
                params = [path_hash]
                cursor.execute(sql, params)
                series_id = cursor.fetchone()[0]

            # series tags
            self._insert_ignore(
                'tags_seriestag',
                ['series_id', 'tag_id', 'value_id'],
                [[series_id, tag_ids[tag], value_ids[value]] for tag, value in parsed.tags.items()]
            )

        return path

    def del_series(self, series, requestContext=None):
        # extract tags and normalize path
        parsed = self.parse(series)
        path_hash = self._path_hash(parsed.path)

        with connection.cursor() as cursor:
            sql = 'SELECT id FROM tags_series WHERE hash=%s'
            params = [path_hash]
            cursor.execute(sql, params)

            row = cursor.fetchone()
            if not row:
                return True

            (series_id, ) = row

            sql = 'DELETE FROM tags_seriestag WHERE series_id=%s'
            params = [series_id]
            cursor.execute(sql, params)

            sql = 'DELETE FROM tags_series WHERE id=%s'
            params = [series_id]
            cursor.execute(sql, params)

        return True



from django.db import models
from graphite.account.models import Profile
from graphite.util import json
import six


class Dashboard(models.Model):
    name = models.CharField(primary_key=True, max_length=128)
    owners = models.ManyToManyField(Profile, related_name='dashboards')
    state = models.TextField()
    __str__ = lambda self: "Dashboard [%s]" % self.name


class Template(models.Model):

    class Admin: pass
    name = models.CharField(primary_key=True, max_length=128)
    owners = models.ManyToManyField(Profile, related_name='templates')
    state = models.TextField()
    __str__ = lambda self: "Template [%s]" % self.name

    def loadState(self, val):
        return self.state.replace('__VALUE__', val)

    def setState(self, state, key):
        #XXX Might not need this
        def replace_string(s):
            if isinstance(s, six.text_type):
                s = s.replace(key, '__VALUE__')
            return s

        def update_graph(graph):
            graph_opts = graph[1]
            graph_opts['target'] = [replace_string(s) for s in graph_opts['target']]
            return [replace_string(graph[0]),
                    graph_opts,
                    replace_string(graph[2])]

        # Parse JSON here and replace first five elements of target with __VALUE__
        parsed_state = json.loads(state)
        for i, graph in enumerate(parsed_state['graphs']):
            parsed_state['graphs'][i] = update_graph(graph)
        self.state = json.dumps(parsed_state)



"""Base tag database"""
import abc
import bisect
import re
import time

from graphite.tags.utils import TaggedSeries


class BaseTagDB(object):
    __metaclass__ = abc.ABCMeta

    def __init__(self, settings, *args, **kwargs):
        """Initialize the tag db."""
        self.settings = settings
        self.cache = kwargs.get('cache')
        self.log = kwargs.get('log')

    def find_series(self, tags, requestContext=None):
        """
        Find series by tag, accepts a list of tag specifiers and returns a list of matching paths.

        Tags specifiers are strings, and may have the following formats:

        .. code-block:: none

          tag=spec    tag value exactly matches spec
          tag!=spec   tag value does not exactly match spec
          tag=~value  tag value matches the regular expression spec
          tag!=~spec  tag value does not match the regular expression spec

        Any tag spec that matches an empty value is considered to match series that don't have that tag.

        At least one tag spec must require a non-empty value.

        Regular expression conditions are treated as being anchored at the start of the value.

        Matching paths are returned as a list of strings.
        """
        start_time = time.time()
        log_msg = 'completed in'

        try:
            cacheKey = self.find_series_cachekey(tags, requestContext=requestContext)
            result = self.cache.get(cacheKey) if self.cache else None
            if result is not None:
                log_msg = 'completed (cached) in'
            else:
                result = self._find_series(tags, requestContext)
                if self.cache:
                    self.cache.set(cacheKey, result, self.settings.TAGDB_CACHE_DURATION)
        except Exception:
            log_msg = 'failed in'
            raise
        finally:
            self.log_info('find_series',
                '{msg} {sec:.6}s'.format(
                    msg=log_msg,
                    sec=time.time() - start_time,
                )
            )

        return result

    def find_series_cachekey(self, tags, requestContext=None):
        return 'TagDB.find_series:' + ':'.join(sorted(tags))

    @abc.abstractmethod
    def _find_series(self, tags, requestContext=None):
        """
        Internal function called by find_series, follows the same semantics allowing base class to implement caching
        """

    @abc.abstractmethod
    def get_series(self, path, requestContext=None):
        """
        Get series by path, accepts a path string and returns a TaggedSeries object describing the series.

        If the path is not found in the TagDB, returns None.
        """

    @abc.abstractmethod
    def list_tags(self, tagFilter=None, limit=None, requestContext=None):
        """
        List defined tags, returns a list of dictionaries describing the tags stored in the TagDB.

        Each tag dict contains the key "tag" which holds the name of the tag.  Additional keys may be returned.

        .. code-block:: none

          [
            {
              'tag': 'tag1',
            },
          ]

        Accepts an optional tagFilter parameter which is a regular expression used to filter the list of returned tags.
        """

    @abc.abstractmethod
    def get_tag(self, tag, valueFilter=None, limit=None, requestContext=None):
        """
        Get details of a particular tag, accepts a tag name and returns a dict describing the tag.

        The dict contains the key "tag" which holds the name of the tag.  It also includes a "values" key,
        which holds a list of the values for each tag.  See list_values() for the structure of each value.

        .. code-block:: none

          {
            'tag': 'tag1',
            'values': [
              {
                'value': 'value1',
                'count': 1,
              }
            ],
          }

        Accepts an optional valueFilter parameter which is a regular expression used to filter the list of returned values.
        """

    @abc.abstractmethod
    def list_values(self, tag, valueFilter=None, limit=None, requestContext=None):
        """
        List values for a particular tag, returns a list of dictionaries describing the values stored in the TagDB.

        Each value dict contains the key "value" which holds the value, and the key "count" which is the number of
        series that have that value.  Additional keys may be returned.

        .. code-block:: none

          [
            {
              'value': 'value1',
              'count': 1,
            },
          ]

        Accepts an optional valueFilter parameter which is a regular expression used to filter the list of returned values.
        """

    @abc.abstractmethod
    def tag_series(self, series, requestContext=None):
        """
        Enter series into database.  Accepts a series string, upserts into the TagDB and returns the canonicalized series name.
        """

    def tag_multi_series(self, seriesList, requestContext=None):
        """
        Enter series into database.  Accepts a list of series strings, upserts into the TagDB and returns a list of canonicalized series names.
        """
        return [self.tag_series(series, requestContext) for series in seriesList]

    @abc.abstractmethod
    def del_series(self, series, requestContext=None):
        """
        Remove series from database.  Accepts a series string and returns True
        """

    def del_multi_series(self, seriesList, requestContext=None):
        """
        Remove series from database.  Accepts a list of series strings, removes them from the TagDB and returns True
        """
        for series in seriesList:
            self.del_series(series, requestContext)
        return True

    def auto_complete_tags(self, exprs, tagPrefix=None, limit=None, requestContext=None):
        """
        Return auto-complete suggestions for tags based on the matches for the specified expressions, optionally filtered by tag prefix
        """
        if limit is None:
            limit = self.settings.TAGDB_AUTOCOMPLETE_LIMIT
        else:
            limit = int(limit)

        if not exprs:
            return [
                tagInfo['tag'] for tagInfo in self.list_tags(
                    tagFilter='^(' + re.escape(tagPrefix) + ')' if tagPrefix else None,
                    limit=limit,
                    requestContext=requestContext,
                )
            ]

        result = []

        searchedTags = set([self.parse_tagspec(expr)[0] for expr in exprs])

        for path in self.find_series(exprs, requestContext=requestContext):
            try:
                tags = self.parse(path).tags
            except Exception:
                continue

            for tag in tags:
                if tag in searchedTags:
                    continue
                if tagPrefix and not tag.startswith(tagPrefix):
                    continue
                if tag in result:
                    continue
                if len(result) == 0 or tag >= result[-1]:
                    if len(result) >= limit:
                        continue
                    result.append(tag)
                else:
                    bisect.insort_left(result, tag)
                if len(result) > limit:
                    del result[-1]

        return result

    def auto_complete_values(self, exprs, tag, valuePrefix=None, limit=None, requestContext=None):
        """
        Return auto-complete suggestions for tags and values based on the matches for the specified expressions, optionally filtered by tag and/or value prefix
        """
        if limit is None:
            limit = self.settings.TAGDB_AUTOCOMPLETE_LIMIT
        else:
            limit = int(limit)

        if not exprs:
            return [
              v['value'] for v in self.list_values(
                tag,
                valueFilter='^(' + re.escape(valuePrefix) + ')' if valuePrefix else None,
                limit=limit,
                requestContext=requestContext,
              )
            ]

        result = []

        for path in self.find_series(exprs + [tag + '!='], requestContext=requestContext):
            try:
                tags = self.parse(path).tags
            except Exception:
                continue

            if tag not in tags:
                continue
            value = tags[tag]
            if valuePrefix and not value.startswith(valuePrefix):
                continue
            if value in result:
                continue
            if len(result) == 0 or value >= result[-1]:
                if len(result) >= limit:
                    continue
                result.append(value)
            else:
                bisect.insort_left(result, value)
            if len(result) > limit:
                del result[-1]

        return result

    def log_info(self, func, msg):
        if self.log:
            self.log.info('%s.%s.%s :: %s' % (self.__module__, self.__class__.__name__, func, msg))

    @staticmethod
    def parse(path):
        return TaggedSeries.parse(path)

    @staticmethod
    def parse_tagspec(tagspec):
        m = re.match('^([^;!=]+)(!?=~?)([^;]*)$', tagspec)
        if m is None:
            raise ValueError("Invalid tagspec %s" % tagspec)

        tag = m.group(1)
        operator = m.group(2)
        spec = m.group(3)

        return (tag, operator, spec)


class DummyTagDB(BaseTagDB):

    def _find_series(self, tags, requestContext=None):
        return []

    def get_series(self, path, requestContext=None):
        return None

    def list_tags(self, tagFilter=None, limit=None, requestContext=None):
        return []

    def get_tag(self, tag, valueFilter=None, limit=None, requestContext=None):
        return None

    def list_values(self, tag, valueFilter=None, limit=None, requestContext=None):
        return []

    def tag_series(self, series, requestContext=None):
        raise NotImplementedError('Tagging not implemented with DummyTagDB')

    def del_series(self, series, requestContext=None):
        return True



#!/usr/bin/env python

import os
import sys
from optparse import OptionParser

from django.core import management

option_parser = OptionParser(usage='''
%prog [options] GRAPHITE_ROOT
''')
option_parser.add_option('--port', default=8080, action='store', type=int, help='Port to listen on')
option_parser.add_option('--interface', default='0.0.0.0', action='store', help='Interface to listen on')
option_parser.add_option('--libs', default=None, help='Path to the directory containing the graphite python package')
option_parser.add_option('--noreload', action='store_true', help='Disable monitoring for changes')
option_parser.add_option('--settings', default='graphite.settings', help='configure alternative settings module')

(options, args) = option_parser.parse_args()

if not args:
    option_parser.print_usage()
    sys.exit(1)

graphite_root = args[0]

python_path = os.path.join(graphite_root, 'webapp')

if options.libs:
    libdir = os.path.expanduser(options.libs)
    print('Adding %s to your PYTHONPATH' % libdir)
    os.environ['PYTHONPATH'] = libdir + ':' + os.environ.get('PYTHONPATH','')

print("Running Graphite from %s under django development server\n" % graphite_root)

command = [
  'django-admin.py',
  'runserver',
  '--pythonpath', python_path,
  '--settings', options.settings,
  '%s:%d' % (options.interface, options.port)
]

if options.noreload:
    command.append('--noreload')

print(' '.join(command))

management.execute_from_command_line(command)



#!/usr/bin/env python

import sys

if sys.version_info <= (2,7):
    # SystemExit defaults to returning 1 when printing a string to stderr
    raise SystemExit("You are using python %s, but version 2.7 or greater is "
                     "required" % sys.version_info)

required = 0
optional = 0


# Test for whisper
try:
    import whisper
except ImportError:
    # No? test for ceres
    try:
        import ceres
        # We imported ceres, but not whisper so it's an optional dependency
        sys.stderr.write("[OPTIONAL] Unable to import the 'whisper' module. "
                         "Without it the webapp will be unable to read .wsp files\n")
        optional += 1
    except ImportError:
        sys.stderr.write("[REQUIRED] Unable to import the 'whisper' or 'ceres' modules, "
                         "please download this package from the Graphite project page and install it.\n")
        required += 1


# Test for cairocffi or pycairo
try:
    import cairocffi as cairo
except ImportError:
    sys.stderr.write("[REQUIRED] Unable to import the 'cairocffi' module, attempting to fall back to pycairo\n")
    try:
        import cairo
    except ImportError:
        sys.stderr.write("[REQUIRED] Unable to import the 'cairo' module, "
                         "do you have pycairo installed for python %s?\n" % sys.version_info.major)
        cairo = None
        required += 1


# Test that pycairo has the PNG backend
try:
    if cairo:
        surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, 10, 10)
        del surface
except Exception:
    sys.stderr.write("[REQUIRED] Failed to create an ImageSurface with cairo, "
                     "you probably need to recompile cairo with PNG support\n")
    required += 1


# Test that cairo can find fonts
try:
    if cairo:
        surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, 10, 10)
        context = cairo.Context(surface)
        context.font_extents()
        del surface, context
except Exception:
    sys.stderr.write("[REQUIRED] Failed to create text with cairo, "
                     "this probably means cairo cant find any fonts. "
                     "Install some system fonts and try again\n")


# Test for django
try:
    import django
except ImportError:
    sys.stderr.write("[REQUIRED] Unable to import the 'django' module, "
                     "do you have Django installed for python %s?\n" % sys.version_info.major)
    django = None
    required += 1


# Test for pytz
try:
    import pytz
except ImportError:
    sys.stderr.write("[REQUIRED] Unable to import the 'pytz' module, "
                     "do you have pytz module installed for python %s?\n" % sys.version_info.major)
    required += 1


# Test for pyparsing
try:
    import pyparsing
except ImportError:
    sys.stderr.write("[REQUIRED] Unable to import the 'pyparsing' module, "
                     "do you have pyparsing module installed for python %s?\n" % sys.version_info.major)
    required += 1


# Test for django-tagging
try:
    import tagging
except ImportError:
    sys.stderr.write("[REQUIRED] Unable to import the 'tagging' module, "
                     "do you have django-tagging installed for python %s?\n" % sys.version_info.major)
    required += 1


if django and django.VERSION[:2] < (1,8):
    sys.stderr.write("[REQUIRED] You have django version %s installed, "
                     "but version 1.8 or greater is required\n" % django.get_version())
    required += 1


# Test for python-memcached
try:
    import memcache
except ImportError:
    sys.stderr.write("[OPTIONAL] Unable to import the 'memcache' module, "
                     "do you have python-memcached installed for python %s? "
                     "This feature is not required but greatly improves performance.\n" % sys.version_info.major)
    optional += 1


# Test for python-ldap
try:
    import ldap
except ImportError:
    sys.stderr.write("[OPTIONAL] Unable to import the 'ldap' module, "
                     "do you have python-ldap installed for python %s? "
                     "Without python-ldap, you will not be able to use "
                     "LDAP authentication in the graphite webapp.\n" % sys.version_info.major)
    optional += 1


# Test for txamqp
try:
    import txamqp
except ImportError:
    sys.stderr.write("[OPTIONAL] Unable to import the 'txamqp' module, "
                     "this is required if you want to use AMQP as an input to Carbon. "
                     "Note that txamqp requires python 2.5 or greater.\n")
    optional += 1


# Test for python-rrdtool
try:
    import rrdtool
except ImportError:
    sys.stderr.write("[OPTIONAL] Unable to import the 'python-rrdtool' module, this is required for reading RRD.\n")
    optional += 1


# Test for whitenoise
try:
    import whitenoise
except ImportError:
    sys.stderr.write("[OPTIONAL] Unable to import the 'whitenoise' module. This is useful for serving static files.\n")
    optional += 1


# Test for pyhash
try:
    import pyhash
except ImportError:
    sys.stderr.write("[OPTIONAL] Unable to import the 'pyhash' module. This is useful for fnv1_ch hashing support.\n")
    optional += 1


if optional:
    sys.stderr.write("%d optional dependencies not met. Please consider the optional items before proceeding.\n" % optional)
else:
    print("All optional dependencies are met.")

if required:
    sys.stderr.write("%d necessary dependencies not met. Graphite will not function until these dependencies are fulfilled.\n" % required)
    sys.exit(1)
else:
    print("All necessary dependencies are met.")


# suppress unused-import warnings
__all__ = [
    'whisper',
    'ceres',
    'cairo',
    'django',
    'pytz',
    'pyparsing',
    'tagging',
    'memcache',
    'ldap',
    'txamqp',
    'rrdtool',
    'whitenoise',
    'pyhash',
]



import inspect

from importlib import import_module
from os import listdir
from os.path import dirname, join, splitext

from django.conf import settings

from graphite.functions.params import Param, ParamTypes, ParamTypeAggFunc  # noqa
from graphite.functions.aggfuncs import getAggFunc  # noqa
from graphite.logger import log

customDir = join(dirname(__file__), 'custom')
customModPrefix = 'graphite.functions.custom.'

_SeriesFunctions = {}
_PieFunctions = {}


def loadFunctions(force=False):
    if _SeriesFunctions and not force:
        return

    from graphite.render import functions

    _SeriesFunctions.clear()
    _SeriesFunctions.update(functions.SeriesFunctions)

    _PieFunctions.clear()
    _PieFunctions.update(functions.PieFunctions)

    custom_modules = []
    for filename in listdir(customDir):
        module_name, extension = splitext(filename)
        if extension != '.py' or module_name == '__init__':
            continue
        custom_modules.append(customModPrefix + module_name)

    for module_name in custom_modules + settings.FUNCTION_PLUGINS:
        try:
            module = import_module(module_name)
        except Exception as e:
            log.warning('Error loading function plugin %s: %s' % (module_name, e))
            continue

        for func_name, func in getattr(module, 'SeriesFunctions', {}).items():
            try:
                addFunction(_SeriesFunctions, func, func_name)
            except Exception as e:
                log.warning('Error loading function plugin %s: %s' % (module_name, e))

        for func_name, func in getattr(module, 'PieFunctions', {}).items():
            try:
                addFunction(_PieFunctions, func, func_name)
            except Exception as e:
                log.warning('Error loading function plugin %s: %s' % (module_name, e))


def addFunction(dest, func, func_name):
    if not hasattr(func, 'group'):
        func.group = 'Ungrouped'

    if not hasattr(func, 'params'):
        raise Exception('No params defined for %s' % func_name)

    for param in func.params:
        if not isinstance(param, Param):
            raise Exception('Invalid param specified for %s' % func_name)
        dest[func_name] = func


def SeriesFunctions():
    loadFunctions()
    return _SeriesFunctions


def SeriesFunction(name):
    loadFunctions()
    try:
        return _SeriesFunctions[name]
    except KeyError:
        raise KeyError('Function "%s" not found' % name)


def PieFunctions():
    loadFunctions()
    return _PieFunctions


def PieFunction(name):
    loadFunctions()
    try:
        return _PieFunctions[name]
    except KeyError:
        raise KeyError('Function "%s" not found' % name)


def functionInfo(name, func):
    argspec = inspect.getargspec(func)
    argformat = inspect.formatargspec(argspec[0][1:], argspec[1], argspec[2], argspec[3])
    return {
        'name': name,
        'function': name + argformat,
        'description': inspect.getdoc(func),
        'module': inspect.getmodule(func).__name__,
        'group': getattr(func, 'group', 'Ungrouped'),
        'params': getattr(func, 'params', None),
    }



# Import some symbols to avoid breaking compatibility.
from graphite.readers.utils import BaseReader, CarbonLink, merge_with_cache  # noqa # pylint: disable=unused-import
from graphite.readers.multi import MultiReader  # noqa # pylint: disable=unused-import
from graphite.readers.whisper import WhisperReader, GzippedWhisperReader  # noqa # pylint: disable=unused-import
from graphite.readers.ceres import CeresReader  # noqa # pylint: disable=unused-import
from graphite.readers.rrd import RRDReader  # noqa # pylint: disable=unused-import



from __future__ import absolute_import

from graphite.intervals import Interval, IntervalSet
from graphite.readers.utils import merge_with_carbonlink, BaseReader

try:
    import ceres
except ImportError:
    ceres = False


class CeresReader(BaseReader):
    __slots__ = ('ceres_node', 'real_metric_path')
    supported = bool(ceres)

    def __init__(self, ceres_node, real_metric_path):
        self.ceres_node = ceres_node
        self.real_metric_path = real_metric_path

    def get_intervals(self):
        intervals = []
        for info in self.ceres_node.slice_info:
            (start, end, step) = info
            intervals.append(Interval(start, end))

        return IntervalSet(intervals)

    def fetch(self, startTime, endTime):
        data = self.ceres_node.read(startTime, endTime)
        time_info = (data.startTime, data.endTime, data.timeStep)
        values = list(data.values)

        values = merge_with_carbonlink(
            self.real_metric_path, data.startTime, data.timeStep, values)

        return time_info, values



from pyparsing import (
    Forward, Combine, Optional, Word, Literal, CaselessKeyword,
    CaselessLiteral, Group, FollowedBy, LineEnd, OneOrMore, ZeroOrMore,
    alphas, alphanums, printables, pyparsing_unicode, delimitedList, quotedString, Regex,
    __version__, Suppress, Empty
)

from django.conf import settings

grammar = Forward()

expression = Forward()

# Literals

intNumber = Regex(r'-?\d+')('integer')

floatNumber = Regex(r'-?\d+\.\d+')('float')

sciNumber = Combine(
  (floatNumber | intNumber) + CaselessLiteral('e') + intNumber
)('scientific')

aString = quotedString('string')

# Use lookahead to match only numbers in a list (can't remember why this is necessary)
afterNumber = FollowedBy(",") ^ FollowedBy(")") ^ FollowedBy(LineEnd())
number = Group(
  (sciNumber + afterNumber) |
  (floatNumber + afterNumber) |
  (intNumber + afterNumber)
)('number')

boolean = Group(
  CaselessKeyword("true") |
  CaselessKeyword("false")
)('boolean')

none = Group(
  CaselessKeyword('none')
)('none')

infinity = Group(
  CaselessKeyword('inf')
)('infinity')

argname = Word(alphas + '_', alphanums + '_')('argname')
funcname = Word(alphas + '_', alphanums + '_')('funcname')

## Symbols
leftParen = Literal('(').suppress()
rightParen = Literal(')').suppress()
comma = Literal(',').suppress()
equal = Literal('=').suppress()

# Function calls

## Symbols
leftBrace = Literal('{')
rightBrace = Literal('}')
leftParen = Literal('(').suppress()
rightParen = Literal(')').suppress()
comma = Literal(',').suppress()
equal = Literal('=').suppress()
backslash = Literal('\\').suppress()

symbols = '''(){},.'"\\|'''
arg = Group(
  boolean |
  number |
  none |
  aString |
  infinity |
  expression
)('args*')
kwarg = Group(argname + equal + arg)('kwargs*')

args = delimitedList(~kwarg + arg)  # lookahead to prevent failing on equals
kwargs = delimitedList(kwarg)


def setRaw(s, loc, toks):
    toks[0]['raw'] = s[toks[0].start:toks[0].end]


call = Group(
  Empty().setParseAction(lambda s, l, t: l)('start') +
  funcname + leftParen +
  Optional(
    args + Optional(
      comma + kwargs
    )
  ) + rightParen +
  Empty().leaveWhitespace().setParseAction(lambda s, l, t: l)('end')
).setParseAction(setRaw)('call')

# Metric pattern (aka. pathExpression)
if settings.UTF8_METRICS:
    Printables = pyparsing_unicode.printables
else:
    Printables = printables
validMetricChars = ''.join((set(Printables) - set(symbols)))
escapedChar = backslash + Word(symbols + '=', exact=1)
partialPathElem = Combine(
  OneOrMore(
    escapedChar | Word(validMetricChars)
  )
)

matchEnum = Combine(
  leftBrace +
  delimitedList(partialPathElem, combine=True) +
  rightBrace
)

pathElement = Combine(
  Group(partialPathElem | matchEnum) +
  ZeroOrMore(matchEnum | partialPathElem)
)
pathExpression = delimitedList(pathElement, delim='.', combine=True)('pathExpression')

litarg = Group(
  number | aString
)('args*')
litkwarg = Group(argname + equal + litarg)('kwargs*')
litargs = delimitedList(~litkwarg + litarg)  # lookahead to prevent failing on equals
litkwargs = delimitedList(litkwarg)

template = Group(
  Literal('template') + leftParen +
  (call | pathExpression) +
  Optional(comma + (litargs | litkwargs)) +
  rightParen
)('template')

pipeSep = ZeroOrMore(Literal(' ')) + Literal('|') + ZeroOrMore(Literal(' '))

pipedExpression = Group(
  (template | call | pathExpression) +
  Group(ZeroOrMore(Suppress(pipeSep) + Group(call)('pipedCall')))('pipedCalls')
)('expression')

if __version__.startswith('1.'):
    expression << pipedExpression
    grammar << expression
else:
    expression <<= pipedExpression
    grammar <<= expression


def enableDebug():
    for name, obj in globals().items():
        try:
            obj.setName(name)
            obj.setDebug(True)
        except Exception:
            pass



import os

from django.db import models

# Monkeypatching so that django-tagging can import python_2_unicode_compatible
import django.utils.encoding
import six
django.utils.encoding.python_2_unicode_compatible = six.python_2_unicode_compatible
django.utils.six = six
from tagging.models import Tag  # noqa: E402

from graphite.events.compat import ModelTaggedItemManager  # noqa: E402

if os.environ.get('READTHEDOCS'):
    TagField = lambda *args, **kwargs: None
else:
    from tagging.fields import TagField


class Event(models.Model):
    when = models.DateTimeField()
    what = models.CharField(max_length=255)
    data = models.TextField(blank=True)
    tags = TagField(default="")

    def get_tags(self):
        return Tag.objects.get_for_object(self)

    def __str__(self):
        return "%s: %s" % (self.when, self.what)

    @staticmethod
    def find_events(time_from=None, time_until=None, tags=None, set_operation=None):

        if tags is not None:
            if set_operation == 'union':
                query = Event.tagged.with_any(tags)
            elif set_operation == 'intersection':
                query = Event.tagged.with_intersection(tags)
            else:
                query = Event.tagged.with_all(tags)
        else:
            query = Event.objects.all()

        if time_from is not None:
            query = query.filter(when__gte=time_from)

        if time_until is not None:
            query = query.filter(when__lte=time_until)

        result = list(query.order_by("when"))
        return result

    def as_dict(self):
        return dict(
            when=self.when,
            what=self.what,
            data=self.data,
            tags=self.tags.split(),
            id=self.id,
        )


# We use this rather than tagging.register() so that tags can be exposed
# in the admin UI
ModelTaggedItemManager().contribute_to_class(Event, 'tagged')




from django import VERSION
from django.http import (HttpResponse as BaseHttpResponse,
                         HttpResponseBadRequest as Base400)


class ContentTypeMixin(object):
    def __init__(self, *args, **kwargs):
        if VERSION < (1, 5) and 'content_type' in kwargs:
            kwargs['mimetype'] = kwargs.pop('content_type')
        super(ContentTypeMixin, self).__init__(*args, **kwargs)


class HttpResponse(ContentTypeMixin, BaseHttpResponse):
    pass


class HttpResponseBadRequest(ContentTypeMixin, Base400):
    pass



"""Copyright 2009 Chris Davis

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License."""

from functools import reduce
import pytz
from six import text_type
from six.moves.urllib.parse import unquote_plus

from datetime import datetime
from django.conf import settings

from graphite.carbonlink import CarbonLink
from graphite.compat import HttpResponse, HttpResponseBadRequest
from graphite.errors import InputParameterError, handleInputParameterError
from graphite.logger import log
from graphite.render.attime import parseATTime
from graphite.storage import STORE, extractForwardHeaders
from graphite.user_util import getProfile
from graphite.util import epoch, json, pickle, msgpack


def index_json(request):
  queryParams = request.GET.copy()
  queryParams.update(request.POST)

  try:
    jsonp = queryParams.get('jsonp', False)

    requestContext = {
      'localOnly': int( queryParams.get('local', 0) ),
      'forwardHeaders': extractForwardHeaders(request),
    }

    matches = STORE.get_index(requestContext)
  except Exception:
    log.exception()
    return json_response_for(request, [], jsonp=jsonp, status=500)

  return json_response_for(request, matches, jsonp=jsonp)


def queryParamAsInt(queryParams, name, default):
  if name not in queryParams:
    return default

  try:
    return int(queryParams[name])
  except Exception as e:
    raise InputParameterError('Invalid int value {value} for param {name}: {err}'.format(
      value=repr(queryParams[name]),
      name=name,
      err=str(e)))


@handleInputParameterError
def find_view(request):
  "View for finding metrics matching a given pattern"

  queryParams = request.GET.copy()
  queryParams.update(request.POST)

  format = queryParams.get('format', 'treejson')
  leaves_only = queryParamAsInt(queryParams, 'leavesOnly', 0)
  local_only = queryParamAsInt(queryParams, 'local', 0)
  wildcards = queryParamAsInt(queryParams, 'wildcards', 0)

  tzinfo = pytz.timezone(settings.TIME_ZONE)
  if 'tz' in queryParams:
    try:
      value = queryParams['tz']
      tzinfo = pytz.timezone(value)
    except pytz.UnknownTimeZoneError:
      pass
    except Exception as e:
      raise InputParameterError(
        'Invalid value {value} for param tz: {err}'
        .format(value=repr(value), err=str(e)))

  if 'now' in queryParams:
    try:
      value = queryParams['now']
      now = parseATTime(value, tzinfo)
    except Exception as e:
      raise InputParameterError(
        'Invalid value {value} for param now: {err}'
        .format(value=repr(value), err=str(e)))
  else:
    now = datetime.now(tzinfo)

  if 'from' in queryParams and str(queryParams['from']) != '-1':
    try:
      value = queryParams['from']
      fromTime = int(epoch(parseATTime(value, tzinfo, now)))
    except Exception as e:
      raise InputParameterError(
        'Invalid value {value} for param from: {err}'
        .format(value=repr(value), err=str(e)))
  else:
    fromTime = -1

  if 'until' in queryParams and str(queryParams['until']) != '-1':
    try:
      value = queryParams['until']
      untilTime = int(epoch(parseATTime(value, tzinfo, now)))
    except Exception as e:
      raise InputParameterError(
        'Invalid value {value} for param until: {err}'
        .format(value=repr(value), err=str(e)))
  else:
    untilTime = -1

  nodePosition = queryParamAsInt(queryParams, 'position', -1)
  jsonp = queryParams.get('jsonp', False)
  forward_headers = extractForwardHeaders(request)

  if fromTime == -1:
    fromTime = None
  if untilTime == -1:
    untilTime = None

  automatic_variants = queryParamAsInt(queryParams, 'automatic_variants', 0)

  try:
    if type(queryParams['query']) is unicode:
      query = queryParams['query'].encode('utf-8')
    else:
      query = str(queryParams['query'])
  except KeyError:
    raise InputParameterError('Missing required parameter \'query\'')
  except NameError:
    query = str(queryParams['query'])

  if query == '':
    raise InputParameterError('Required parameter \'query\' is empty')

  if '.' in query:
    base_path = query.rsplit('.', 1)[0] + '.'
  else:
    base_path = ''

  if format == 'completer':
    query = query.replace('..', '*.')
    if not query.endswith('*'):
      query += '*'

    if automatic_variants:
      query_parts = query.split('.')
      for i,part in enumerate(query_parts):
        if ',' in part and '{' not in part:
          query_parts[i] = '{%s}' % part
      query = '.'.join(query_parts)

  try:
    matches = list(STORE.find(
      query, fromTime, untilTime,
      local=local_only,
      headers=forward_headers,
      leaves_only=leaves_only,
    ))
  except Exception:
    log.exception()
    raise

  log.info('find_view query=%s local_only=%s matches=%d' % (query, local_only, len(matches)))
  matches.sort(key=lambda node: node.name)
  log.info("received remote find request: pattern=%s from=%s until=%s local_only=%s format=%s matches=%d" % (query, fromTime, untilTime, local_only, format, len(matches)))

  if format == 'treejson':
    profile = getProfile(request)
    content = tree_json(matches, base_path, wildcards=profile.advancedUI or wildcards)
    response = json_response_for(request, content, jsonp=jsonp)

  elif format == 'nodelist':
    content = nodes_by_position(matches, nodePosition)
    response = json_response_for(request, content, jsonp=jsonp)

  elif format == 'pickle':
    content = pickle_nodes(matches)
    response = HttpResponse(content, content_type='application/pickle')

  elif format == 'msgpack':
    content = msgpack_nodes(matches)
    response = HttpResponse(content, content_type='application/x-msgpack')

  elif format == 'json':
    content = json_nodes(matches)
    response = json_response_for(request, content, jsonp=jsonp)

  elif format == 'completer':
    results = []
    for node in matches:
      node_info = dict(path=node.path, name=node.name, is_leaf=str(int(node.is_leaf)))
      if not node.is_leaf:
        node_info['path'] += '.'
      results.append(node_info)

    if len(results) > 1 and wildcards:
      wildcardNode = {'name' : '*'}
      results.append(wildcardNode)

    response = json_response_for(request, { 'metrics' : results }, jsonp=jsonp)

  else:
    return HttpResponseBadRequest(
        content="Invalid value for 'format' parameter",
        content_type='text/plain')

  response['Pragma'] = 'no-cache'
  response['Cache-Control'] = 'no-cache'
  return response


def expand_view(request):
  "View for expanding a pattern into matching metric paths"
  queryParams = request.GET.copy()
  queryParams.update(request.POST)

  local_only = int( queryParams.get('local', 0) )
  group_by_expr = int( queryParams.get('groupByExpr', 0) )
  leaves_only = int( queryParams.get('leavesOnly', 0) )
  jsonp = queryParams.get('jsonp', False)
  forward_headers = extractForwardHeaders(request)

  results = {}
  for query in queryParams.getlist('query'):
    results[query] = set()
    for node in STORE.find(query, local=local_only, headers=forward_headers):
      if node.is_leaf or not leaves_only:
        results[query].add( node.path )

  # Convert our results to sorted lists because sets aren't json-friendly
  if group_by_expr:
    for query, matches in results.items():
      results[query] = sorted(matches)
  else:
    results = sorted( reduce(set.union, results.values(), set()) )

  result = {
    'results' : results
  }

  response = json_response_for(request, result, jsonp=jsonp)
  response['Pragma'] = 'no-cache'
  response['Cache-Control'] = 'no-cache'
  return response


def get_metadata_view(request):
  queryParams = request.GET.copy()
  queryParams.update(request.POST)

  key = queryParams.get('key')
  metrics = queryParams.getlist('metric')
  jsonp = queryParams.get('jsonp', False)
  results = {}
  for metric in metrics:
    try:
      results[metric] = CarbonLink.get_metadata(metric, key)
    except Exception:
      log.exception()
      results[metric] = dict(error="Unexpected error occurred in CarbonLink.get_metadata(%s, %s)" % (metric, key))

  return json_response_for(request, results, jsonp=jsonp)


def set_metadata_view(request):
  results = {}

  if request.method == 'GET':
    metric = request.GET['metric']
    key = request.GET['key']
    value = request.GET['value']
    try:
      results[metric] = CarbonLink.set_metadata(metric, key, value)
    except Exception:
      log.exception()
      results[metric] = dict(error="Unexpected error occurred in CarbonLink.set_metadata(%s, %s)" % (metric, key))

  elif request.method == 'POST':
    if request.META.get('CONTENT_TYPE') == 'application/json':
      operations = json.loads( request.body )
    else:
      operations = json.loads( request.POST['operations'] )

    for op in operations:
      metric = None
      try:
        metric, key, value = op['metric'], op['key'], op['value']
        results[metric] = CarbonLink.set_metadata(metric, key, value)
      except Exception:
        log.exception()
        if metric:
          results[metric] = dict(error="Unexpected error occurred in bulk CarbonLink.set_metadata(%s)" % metric)

  else:
    results = dict(error='Invalid request method')

  return json_response_for(request, results)


def tree_json(nodes, base_path, wildcards=False):
  results = []

  branchNode = {
    'allowChildren': 1,
    'expandable': 1,
    'leaf': 0,
  }
  leafNode = {
    'allowChildren': 0,
    'expandable': 0,
    'leaf': 1,
  }

  #Add a wildcard node if appropriate
  if len(nodes) > 1 and wildcards:
    wildcardNode = {'text' : '*', 'id' : base_path + '*'}

    if any(not n.is_leaf for n in nodes):
      wildcardNode.update(branchNode)

    else:
      wildcardNode.update(leafNode)

    results.append(wildcardNode)

  found = set()
  results_leaf = []
  results_branch = []
  for node in nodes: #Now let's add the matching children
    if node.name in found:
      continue

    found.add(node.name)
    resultNode = {
      'text' : unquote_plus(str(node.name)),
      'id' : base_path + str(node.name),
    }

    if node.is_leaf:
      resultNode.update(leafNode)
      results_leaf.append(resultNode)
    else:
      resultNode.update(branchNode)
      results_branch.append(resultNode)

  results.extend(results_branch)
  results.extend(results_leaf)
  return results


def nodes_by_position(matches, position):
  found = set()

  for metric in matches:
    nodes = metric.path.split('.')
    found.add(nodes[position])
  results = { 'nodes' : sorted(found) }
  return results


def pickle_nodes(nodes):
  nodes_info = []

  for node in nodes:
    info = dict(path=node.path, is_leaf=node.is_leaf)
    if node.is_leaf:
      info['intervals'] = node.intervals

    nodes_info.append(info)

  return pickle.dumps(nodes_info, protocol=-1)


def msgpack_nodes(nodes):
  nodes_info = []

  # make sure everything is unicode in python 2.x and 3.x
  for node in nodes:
    info = {
      text_type('path'): text_type(node.path),
      text_type('is_leaf'): node.is_leaf,
    }
    if node.is_leaf:
      info[text_type('intervals')] = [interval.tuple for interval in node.intervals]

    nodes_info.append(info)

  return msgpack.dumps(nodes_info, use_bin_type=True)


def json_nodes(nodes):
  nodes_info = []

  for node in nodes:
    info = dict(path=node.path, is_leaf=node.is_leaf)
    if node.is_leaf:
      info['intervals'] = [{'start': i.start, 'end': i.end} for i in node.intervals]

    nodes_info.append(info)

  return sorted(nodes_info, key=lambda item: item['path'])


def json_response_for(request, data, content_type='application/json', jsonp=False, **kwargs):
  accept = request.META.get('HTTP_ACCEPT', 'application/json')
  ensure_ascii = accept == 'application/json'

  pretty = bool(request.POST.get('pretty', request.GET.get('pretty')))

  content = json.dumps(data, ensure_ascii=ensure_ascii, indent=(2 if pretty else None))
  if jsonp:
    content = "%s(%s)" % (jsonp, content)
    content_type = 'text/javascript'
  if not ensure_ascii:
    content_type += ';charset=utf-8'

  return HttpResponse(content, content_type=content_type, **kwargs)



import re
import errno

from os.path import getmtime
from six.moves.urllib.parse import urlencode
from six.moves.configparser import ConfigParser
from django.shortcuts import render
from django.http import QueryDict
from django.conf import settings
from django.contrib.auth import login, authenticate, logout
from django.contrib.staticfiles import finders
from django.utils.safestring import mark_safe
from graphite.compat import HttpResponse
from graphite.dashboard.models import Dashboard, Template
from graphite.dashboard.send_graph import send_graph_email
from graphite.render.views import renderView
from graphite.util import json
from graphite.user_util import isAuthenticated

fieldRegex = re.compile(r'<([^>]+)>')
defaultScheme = {
  'name' : 'Everything',
  'pattern' : '<category>',
  'fields' : [ dict(name='category', label='Category') ],
}
defaultUIConfig = {
  'default_graph_width'  : 400,
  'default_graph_height' : 250,
  'refresh_interval'     :  60,
  'autocomplete_delay'   : 375,
  'merge_hover_delay'    : 700,
  'theme'                : 'default',
}
defaultKeyboardShortcuts = {
  'toggle_toolbar' : 'ctrl-z',
  'toggle_metrics_panel' : 'ctrl-space',
  'erase_all_graphs' : 'alt-x',
  'save_dashboard' : 'alt-s',
  'completer_add_metrics' : 'alt-enter',
  'completer_del_metrics' : 'alt-backspace',
  'give_completer_focus' : 'shift-space',
}

ALL_PERMISSIONS = ['change', 'delete']


class DashboardConfig:
  def __init__(self):
    self.last_read = 0
    self.schemes = [defaultScheme]
    self.ui_config = defaultUIConfig.copy()

  def check(self):
    if getmtime(settings.DASHBOARD_CONF) > self.last_read:
      self.load()

  def load(self):
    schemes = [defaultScheme]
    parser = ConfigParser()
    parser.read(settings.DASHBOARD_CONF)

    for option, default_value in defaultUIConfig.items():
      if parser.has_option('ui', option):
        try:
          self.ui_config[option] = parser.getint('ui', option)
        except ValueError:
          self.ui_config[option] = parser.get('ui', option)
      else:
        self.ui_config[option] = default_value

    if parser.has_option('ui', 'automatic_variants'):
      self.ui_config['automatic_variants']   = parser.getboolean('ui', 'automatic_variants')
    else:
      self.ui_config['automatic_variants'] = True

    self.ui_config['keyboard_shortcuts'] = defaultKeyboardShortcuts.copy()
    if parser.has_section('keyboard-shortcuts'):
      self.ui_config['keyboard_shortcuts'].update( parser.items('keyboard-shortcuts') )

    for section in parser.sections():
      if section in ('ui', 'keyboard-shortcuts'):
        continue

      scheme = parser.get(section, 'scheme')
      fields = []

      for match in fieldRegex.finditer(scheme):
        field = match.group(1)
        if parser.has_option(section, '%s.label' % field):
          label = parser.get(section, '%s.label' % field)
        else:
          label = field

        fields.append({
          'name' : field,
          'label' : label
        })

      schemes.append({
        'name' : section,
        'pattern' : scheme,
        'fields' : fields,
      })

    self.schemes = schemes


config = DashboardConfig()


def dashboard(request, name=None):
  dashboard_conf_missing = False

  try:
    config.check()
  except OSError as e:
    if e.errno == errno.ENOENT:
      dashboard_conf_missing = True
    else:
      raise

  initialError = None
  debug = request.GET.get('debug', False)
  theme = request.GET.get('theme', config.ui_config['theme'])
  css_file = finders.find('css/dashboard-%s.css' % theme)
  if css_file is None:
    initialError = "Invalid theme '%s'" % theme
    theme = config.ui_config['theme']

  context = {
    'schemes_json': mark_safe(json.dumps(config.schemes)),
    'ui_config_json': mark_safe(json.dumps(config.ui_config)),
    'jsdebug': debug or settings.JAVASCRIPT_DEBUG,
    'debug': debug,
    'theme': theme,
    'initialError': initialError,
    'querystring': mark_safe(json.dumps(dict(request.GET.items()))),
    'dashboard_conf_missing': dashboard_conf_missing,
    'userName': '',
    'permissions': mark_safe(json.dumps(getPermissions(request.user))),
    'permissionsUnauthenticated': mark_safe(json.dumps(getPermissions(None)))
  }
  user = request.user
  if user:
      context['userName'] = user.username

  if name is not None:
    try:
      dashboard = Dashboard.objects.get(name=name)
    except Dashboard.DoesNotExist:
      context['initialError'] = "Dashboard '%s' does not exist." % name
    else:
      context['initialState'] = dashboard.state

  return render(request, "dashboard.html", context)


def template(request, name, val):
  template_conf_missing = False

  try:
    config.check()
  except OSError as e:
    if e.errno == errno.ENOENT:
      template_conf_missing = True
    else:
      raise

  initialError = None
  debug = request.GET.get('debug', False)
  theme = request.GET.get('theme', config.ui_config['theme'])
  css_file = finders.find('css/dashboard-%s.css' % theme)
  if css_file is None:
    initialError = "Invalid theme '%s'" % theme
    theme = config.ui_config['theme']

  context = {
    'schemes_json' : json.dumps(config.schemes),
    'ui_config_json' : json.dumps(config.ui_config),
    'jsdebug' : debug or settings.JAVASCRIPT_DEBUG,
    'debug' : debug,
    'theme' : theme,
    'initialError' : initialError,
    'querystring' : json.dumps( dict( request.GET.items() ) ),
    'template_conf_missing' : template_conf_missing,
    'userName': '',
    'permissions': json.dumps(getPermissions(request.user)),
    'permissionsUnauthenticated': json.dumps(getPermissions(None))
  }

  user = request.user
  if user:
      context['userName'] = user.username

  try:
    template = Template.objects.get(name=name)
  except Template.DoesNotExist:
    context['initialError'] = "Template '%s' does not exist." % name
  else:
    state = json.loads(template.loadState(val))
    state['name'] = '%s/%s' % (name, val)
    context['initialState'] = json.dumps(state)
  return render(request, "dashboard.html", context)


def getPermissions(user):
  """Return [change, delete] based on authorisation model and user privileges/groups"""
  if user and not isAuthenticated(user):
    user = None
  if not settings.DASHBOARD_REQUIRE_AUTHENTICATION:
    return ALL_PERMISSIONS      # don't require login
  if not user:
      return []
  # from here on, we have a user
  permissions = ALL_PERMISSIONS
  if settings.DASHBOARD_REQUIRE_PERMISSIONS:
    permissions = [permission for permission in ALL_PERMISSIONS if user.has_perm('dashboard.%s_dashboard' % permission)]
  editGroup = settings.DASHBOARD_REQUIRE_EDIT_GROUP
  if editGroup and len(user.groups.filter(name = editGroup)) == 0:
    permissions = []
  return permissions


def save(request, name):
  if 'change' not in getPermissions(request.user):
    return json_response( dict(error="Must be logged in with appropriate permissions to save") )
  # Deserialize and reserialize as a validation step
  state = str( json.dumps( json.loads( request.POST['state'] ) ) )

  try:
    dashboard = Dashboard.objects.get(name=name)
  except Dashboard.DoesNotExist:
    dashboard = Dashboard.objects.create(name=name, state=state)
  else:
    dashboard.state = state
    dashboard.save()

  return json_response( dict(success=True) )


def save_template(request, name, key):
  if 'change' not in getPermissions(request.user):
    return json_response( dict(error="Must be logged in with appropriate permissions to save the template") )
  # Deserialize and reserialize as a validation step
  state = str( json.dumps( json.loads( request.POST['state'] ) ) )

  try:
    template = Template.objects.get(name=name)
  except Template.DoesNotExist:
    template = Template.objects.create(name=name)
    template.setState(state, key)
    template.save()
  else:
    template.setState(state, key)
    template.save()

  return json_response( dict(success=True) )


def load(request, name):
  try:
    dashboard = Dashboard.objects.get(name=name)
  except Dashboard.DoesNotExist:
    return json_response( dict(error="Dashboard '%s' does not exist. " % name) )

  return json_response( dict(state=json.loads(dashboard.state)) )


def load_template(request, name, val):
  try:
    template = Template.objects.get(name=name)
  except Template.DoesNotExist:
    return json_response( dict(error="Template '%s' does not exist. " % name) )

  state = json.loads(template.loadState(val))
  state['name'] = '%s/%s' % (name, val)
  return json_response( dict(state=state) )


def delete(request, name):
  if 'delete' not in getPermissions(request.user):
    return json_response( dict(error="Must be logged in with appropriate permissions to delete") )

  try:
    dashboard = Dashboard.objects.get(name=name)
  except Dashboard.DoesNotExist:
    return json_response( dict(error="Dashboard '%s' does not exist. " % name) )
  else:
    dashboard.delete()
    return json_response( dict(success=True) )


def delete_template(request, name):
  if 'delete' not in getPermissions(request.user):
    return json_response( dict(error="Must be logged in with appropriate permissions to delete the template") )

  try:
    template = Template.objects.get(name=name)
  except Template.DoesNotExist:
    return json_response( dict(error="Template '%s' does not exist. " % name) )
  else:
    template.delete()
    return json_response( dict(success=True) )


def find(request):
  queryParams = request.GET.copy()
  queryParams.update(request.POST)

  query = queryParams.get('query', False)
  query_terms = set( query.lower().split() )
  results = []

  # Find all dashboard names that contain each of our query terms as a substring
  for dashboard_name in Dashboard.objects.order_by('name').values_list('name', flat=True):
    name = dashboard_name.lower()

    if name.startswith('temporary-'):
      continue

    found = True # blank queries return everything
    for term in query_terms:
      if term in name:
        found = True
      else:
        found = False
        break

    if found:
      results.append( dict(name=dashboard_name) )

  return json_response( dict(dashboards=results) )


def find_template(request):
  queryParams = request.GET.copy()
  queryParams.update(request.POST)

  query = queryParams.get('query', False)
  query_terms = set( query.lower().split() )
  results = []

  # Find all dashboard names that contain each of our query terms as a substring
  for template in Template.objects.all():
    name = template.name.lower()

    found = True # blank queries return everything
    for term in query_terms:
      if term in name:
        found = True
      else:
        found = False
        break

    if found:
      results.append( dict(name=template.name) )

  return json_response( dict(templates=results) )


def help(request):
  context = {}
  return render(request, "dashboardHelp.html", context)


def email(request):
    sender = request.POST['sender']
    recipients = request.POST['recipients'].split()
    subject = request.POST['subject']
    message = request.POST['message']

    # these need to be passed to the render function in an HTTP request.
    graph_params = json.loads(request.POST['graph_params'], parse_int=str)
    target = QueryDict(urlencode({'target': graph_params.pop('target')}))
    graph_params = QueryDict(urlencode(graph_params))

    new_post = request.POST.copy()
    new_post.update(graph_params)
    new_post.update(target)
    request.POST = new_post

    resp = renderView(request)
    img = resp.content

    if img:
        attachments = [('graph.png', img, 'image/png')]
        send_graph_email(subject, sender, recipients, attachments, message)

    return json_response(dict(success=True))


def create_temporary(request):
  state = str( json.dumps( json.loads( request.POST['state'] ) ) )
  i = 0
  while True:
    name = "temporary-%d" % i
    try:
      Dashboard.objects.get(name=name)
    except Dashboard.DoesNotExist:
      dashboard = Dashboard.objects.create(name=name, state=state)
      break
    else:
      i += 1

  return json_response( dict(name=dashboard.name) )


def json_response(obj):
  return HttpResponse(content_type='application/json', content=json.dumps(obj))


def user_login(request):
  response = dict(errors={}, text={}, success=False, permissions=[])
  user = authenticate(username=request.POST['username'],
                      password=request.POST['password'])
  if user is not None:
    if user.is_active:
      login(request, user)
      response['success'] = True
      response['permissions'].extend(getPermissions(user))
    else:
      response['errors']['reason'] = 'Account disabled.'
  else:
    response['errors']['reason'] = 'Username and/or password invalid.'
  return json_response(response)


def user_logout(request):
  response = dict(errors={}, text={}, success=True)
  logout(request)
  return json_response(response)



"""Copyright 2008 Orbitz WorldWide
   Copyright 2011 Chris Davis

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License."""

from hashlib import md5
from itertools import chain
import bisect
import sys

try:
    import mmh3
except ImportError:
    mmh3 = None

try:
    import pyhash
    hasher = pyhash.fnv1a_32()

    def fnv32a(data, seed=0x811c9dc5):
        return hasher(data, seed=seed)
except ImportError:
    def fnv32a(data, seed=0x811c9dc5):
        """
        FNV-1a Hash (http://isthe.com/chongo/tech/comp/fnv/) in Python.
        Taken from https://gist.github.com/vaiorabbit/5670985
        """
        hval = seed
        fnv_32_prime = 0x01000193
        uint32_max = 2 ** 32
        if sys.version_info >= (3, 0):
            # data is a bytes object, s is an integer
            for s in data:
                hval = hval ^ s
                hval = (hval * fnv_32_prime) % uint32_max
        else:
            # data is an str object, s is a single character
            for s in data:
                hval = hval ^ ord(s)
                hval = (hval * fnv_32_prime) % uint32_max
        return hval


def hashRequest(request):
    # Normalize the request parameters to ensure we're deterministic
    queryParams = [
        "%s=%s" % (key, '&'.join(values))
        for (key,values) in chain(request.POST.lists(), request.GET.lists())
        if not key.startswith('_')
    ]
    normalizedParams = ','.join( sorted(queryParams) )
    return compactHash(normalizedParams)


def hashData(targets, startTime, endTime, xFilesFactor):
    targetsString = ','.join(sorted(targets))
    startTimeString = startTime.strftime("%Y%m%d_%H%M")
    endTimeString = endTime.strftime("%Y%m%d_%H%M")
    myHash = targetsString + '@' + startTimeString + ':' + endTimeString + ':' + str(xFilesFactor)
    return compactHash(myHash)


def compactHash(string):
    return md5(string.encode('utf-8')).hexdigest()


def carbonHash(key, hash_type):
    if hash_type == 'fnv1a_ch':
        big_hash = int(fnv32a(key.encode('utf-8')))
        small_hash = (big_hash >> 16) ^ (big_hash & 0xffff)
    elif hash_type == 'mmh3_ch':
        if mmh3 is None:
            raise Exception('Install "mmh3" to use this hashing function.')
        small_hash = mmh3.hash(key)
    else:
        big_hash = compactHash(key)
        small_hash = int(big_hash[:4], 16)
    return small_hash


class ConsistentHashRing:
    def __init__(self, nodes, replica_count=100, hash_type='carbon_ch'):
        self.ring = []
        self.ring_len = len(self.ring)
        self.nodes = set()
        self.nodes_len = len(self.nodes)
        self.replica_count = replica_count
        self.hash_type = hash_type
        for node in nodes:
            self.add_node(node)

    def compute_ring_position(self, key):
        return carbonHash(key, self.hash_type)

    def add_node(self, key):
        self.nodes.add(key)
        self.nodes_len = len(self.nodes)
        for i in range(self.replica_count):
            if self.hash_type == 'fnv1a_ch':
                replica_key = "%d-%s" % (i, key[1])
            else:
                replica_key = "%s:%d" % (key, i)
            position = self.compute_ring_position(replica_key)
            while position in [r[0] for r in self.ring]:
                position = position + 1
            entry = (position, key)
            bisect.insort(self.ring, entry)
        self.ring_len = len(self.ring)

    def remove_node(self, key):
        self.nodes.discard(key)
        self.nodes_len = len(self.nodes)
        self.ring = [entry for entry in self.ring if entry[1] != key]
        self.ring_len = len(self.ring)

    def get_node(self, key):
        assert self.ring
        position = self.compute_ring_position(key)
        search_entry = (position, ())
        index = bisect.bisect_left(self.ring, search_entry) % self.ring_len
        entry = self.ring[index]
        return entry[1]

    def get_nodes(self, key):
        nodes = set()
        if not self.ring:
            return
        if self.nodes_len == 1:
            for node in self.nodes:
                yield node
        position = self.compute_ring_position(key)
        search_entry = (position, ())
        index = bisect.bisect_left(self.ring, search_entry) % self.ring_len
        last_index = (index - 1) % self.ring_len
        nodes_len = len(nodes)
        while nodes_len < self.nodes_len and index != last_index:
            next_entry = self.ring[index]
            (position, next_node) = next_entry
            if next_node not in nodes:
                nodes.add(next_node)
                nodes_len += 1
                yield next_node

            index = (index + 1) % self.ring_len



"""Copyright 2008 Orbitz WorldWide

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License."""


from django import VERSION as DJANGO_VERSION
from django.contrib.auth.models import User
from graphite.account.models import Profile
from graphite.logger import log


def isAuthenticated(user):
    # is_authenticated() is changed to a boolean since 1.10, 2.0 removes the
    # backwards compatibilty
    if DJANGO_VERSION >= (1, 10):
        return user.is_authenticated
    else:
        return user.is_authenticated()


def getProfile(request, allowDefault=True):
    if isAuthenticated(request.user):
        return Profile.objects.get_or_create(user=request.user)[0]
    elif allowDefault:
        return default_profile()


def getProfileByUsername(username):
    try:
        return Profile.objects.get(user__username=username)
    except Profile.DoesNotExist:
        return None


def default_profile():
    # '!' is an unusable password. Since the default user never authenticates
    # this avoids creating a default (expensive!) password hash at every
    # default_profile() call.
    user, created = User.objects.get_or_create(
        username='default', defaults={'email': 'default@localhost.localdomain',
                                      'password': '!'})
    if created:
        log.info("Default user didn't exist, created it")
    profile, created = Profile.objects.get_or_create(user=user)
    if created:
        log.info("Default profile didn't exist, created it")
    return profile



"""Copyright 2008 Orbitz WorldWide

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License."""

import imp
import io
import json as _json
import socket
import time
import sys
import calendar
import pytz
import six
import traceback

from datetime import datetime
from functools import wraps
from os.path import splitext, basename

from django.conf import settings
from django.utils.timezone import make_aware

from graphite.compat import HttpResponse
from graphite.logger import log

# BytesIO is needed on py3 as StringIO does not operate on byte input anymore
# We could use BytesIO on py2 as well but it is slower than StringIO
if sys.version_info >= (3, 0):
  PY3 = True
  import pickle
  from io import BytesIO
else:
  PY3 = False
  import cPickle as pickle
  from cStringIO import StringIO as BytesIO

# use https://github.com/msgpack/msgpack-python if available
try:
  import msgpack  # NOQA
# otherwise fall back to bundled https://github.com/vsergeev/u-msgpack-python
except ImportError:
  import graphite.umsgpack as msgpack  # NOQA


def epoch(dt):
  """
  Returns the epoch timestamp of a timezone-aware datetime object.
  """
  if not dt.tzinfo:
    tb = traceback.extract_stack(None, 2)
    log.warning('epoch() called with non-timezone-aware datetime in %s at %s:%d' % (tb[0][2], tb[0][0], tb[0][1]))
    return calendar.timegm(make_aware(dt, pytz.timezone(settings.TIME_ZONE)).astimezone(pytz.utc).timetuple())
  return calendar.timegm(dt.astimezone(pytz.utc).timetuple())


def epoch_to_dt(timestamp):
    """
    Returns the timezone-aware datetime of an epoch timestamp.
    """
    return make_aware(datetime.utcfromtimestamp(timestamp), pytz.utc)


def timebounds(requestContext):
  startTime = int(epoch(requestContext['startTime']))
  endTime = int(epoch(requestContext['endTime']))
  now = int(epoch(requestContext['now']))

  return (startTime, endTime, now)


def is_local_interface(host):
  is_ipv6 = False
  if ':' not in host:
    pass
  elif host.count(':') == 1:
    host = host.split(':', 1)[0]
  else:
    is_ipv6 = True

    if host.find('[', 0, 2) != -1:
      last_bracket_position  = host.rfind(']')
      last_colon_position = host.rfind(':')
      if last_colon_position > last_bracket_position:
        host = host.rsplit(':', 1)[0]
      host = host.strip('[]')

  try:
    if is_ipv6:
      sock = socket.socket(socket.AF_INET6, socket.SOCK_DGRAM)
    else:
      sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    sock.bind( (host, 0) )
  except socket.error:
    return False
  finally:
    sock.close()

  return True


def is_pattern(s):
   return '*' in s or '?' in s or '[' in s or '{' in s


def is_escaped_pattern(s):
  for symbol in '*?[{':
    i = s.find(symbol)
    if i > 0:
      if s[i-1] == '\\':
        return True
  return False


def find_escaped_pattern_fields(pattern_string):
  pattern_parts = pattern_string.split('.')
  for index,part in enumerate(pattern_parts):
    if is_escaped_pattern(part):
      yield index


def load_module(module_path, member=None):
  module_name = splitext(basename(module_path))[0]
  try:  # 'U' is default from Python 3.0 and deprecated since 3.9
    module_file = open(module_path, 'U')
  except ValueError:
    module_file = open(module_path, 'rt')
  description = ('.py', 'U', imp.PY_SOURCE)
  module = imp.load_module(module_name, module_file, module_path, description)
  if member:
    return getattr(module, member)
  else:
    return module


def timestamp(dt):
  "Convert a datetime object into epoch time"
  return time.mktime(dt.timetuple())


def deltaseconds(timedelta):
  "Convert a timedelta object into seconds (same as timedelta.total_seconds() in Python 2.7+)"
  return (timedelta.microseconds + (timedelta.seconds + timedelta.days * 24 * 3600) * 10**6) / 10**6


# This whole song & dance is due to pickle being insecure
# The SafeUnpickler classes were largely derived from
# http://nadiana.com/python-pickle-insecure
# This code also lives in carbon.util
if not PY3:
  class SafeUnpickler(object):
    PICKLE_SAFE = {
      'copy_reg': set(['_reconstructor']),
      '__builtin__': set(['object', 'list', 'set']),
      'collections': set(['deque']),
      'graphite.render.datalib': set(['TimeSeries', 'Tags']),
      'graphite.intervals': set(['Interval', 'IntervalSet']),
    }

    @classmethod
    def find_class(cls, module, name):
      if module not in cls.PICKLE_SAFE:
        raise pickle.UnpicklingError('Attempting to unpickle unsafe module %s' % module)
      __import__(module)
      mod = sys.modules[module]
      if name not in cls.PICKLE_SAFE[module]:
        raise pickle.UnpicklingError('Attempting to unpickle unsafe class %s' % name)
      return getattr(mod, name)

    @classmethod
    def loads(cls, pickle_string):
      pickle_obj = pickle.Unpickler(BytesIO(pickle_string))
      pickle_obj.find_global = cls.find_class
      return pickle_obj.load()

    @classmethod
    def load(cls, file):
      pickle_obj = pickle.Unpickler(file)
      pickle_obj.find_global = cls.find_class
      return pickle_obj.load()

  unpickle = SafeUnpickler

else:
  class SafeUnpickler(pickle.Unpickler):
    PICKLE_SAFE = {
      'copy_reg': set(['_reconstructor']),
      'builtins': set(['object', 'list', 'set']),
      'collections': set(['deque']),
      'graphite.render.datalib': set(['TimeSeries', 'Tags']),
      'graphite.intervals': set(['Interval', 'IntervalSet']),
    }

    def __init__(self, file):
        super().__init__(file, encoding='utf8')

    def find_class(self, module, name):
      if module not in self.PICKLE_SAFE:
        raise pickle.UnpicklingError('Attempting to unpickle unsafe module %s' % module)
      __import__(module)
      mod = sys.modules[module]
      if name not in self.PICKLE_SAFE[module]:
        raise pickle.UnpicklingError('Attempting to unpickle unsafe class %s' % name)
      return getattr(mod, name)

  class unpickle(object):
    @staticmethod
    def loads(pickle_string):
      return SafeUnpickler(BytesIO(pickle_string)).load()

    @staticmethod
    def load(file):
      return SafeUnpickler(file).load()


class json(object):
  JSONEncoder = _json.JSONEncoder
  JSONDecoder = _json.JSONDecoder

  @staticmethod
  def dump(*args, **kwargs):
    return _json.dump(*args, **kwargs)

  @staticmethod
  def dumps(*args, **kwargs):
    return _json.dumps(*args, **kwargs)

  @staticmethod
  def load(fp, *args, **kwargs):
    return _json.load(fp, *args, **kwargs)

  @staticmethod
  def loads(s, *args, **kwargs):
    if isinstance(s, six.binary_type):
      return _json.loads(s.decode('utf-8'), *args, **kwargs)
    return _json.loads(s, *args, **kwargs)


class Timer(object):
  __slots__ = ('msg', 'name', 'start_time')

  def __init__(self, name):
    self.name = name
    self.msg = 'completed in'
    self.start_time = time.time()

  def set_msg(self, msg):
    self.msg = msg

  def set_name(self, name):
    self.name = name

  def stop(self):
    log.info(
      '{name} :: {msg} {sec:.6}s'.format(
        name=self.name,
        msg=self.msg,
        sec=time.time() - self.start_time,
      )
    )


def logtime(f):
  @wraps(f)
  def wrapped_f(*args, **kwargs):
    timer = Timer(f.__module__ + '.' + f.__name__)
    kwargs['timer'] = timer

    try:
      return f(*args, **kwargs)
    except Exception:
      timer.msg = 'failed in'
      raise
    finally:
      timer.stop()

  return wrapped_f


class BufferedHTTPReader(io.FileIO):
  def __init__(self, response, buffer_size=1048576):
    self.response = response
    self.buffer_size = buffer_size
    self.buffer = b''
    self.pos = 0

  def read(self, amt=None):
    if amt is None:
      return self.response.read()
    if len(self.buffer) - self.pos < amt:
      self.buffer = self.buffer[self.pos:]
      self.pos = 0
      self.buffer += self.response.read(self.buffer_size)
    data = self.buffer[self.pos:self.pos + amt]
    self.pos += amt
    if self.pos >= len(self.buffer):
      self.pos = 0
      self.buffer = b''
    return data


def jsonResponse(*args, **kwargs):
  encoder = kwargs.get('encoder')
  default = kwargs.get('default')

  def decorator(f):
    @wraps(f)
    def wrapped_f(request, *args, **kwargs):
      if request.method == 'GET':
        queryParams = request.GET.copy()
      elif request.method == 'POST':
        queryParams = request.GET.copy()
        queryParams.update(request.POST)
      else:
        queryParams = {}

      try:
        return _jsonResponse(
          f(request, queryParams, *args, **kwargs), queryParams, encoder=encoder, default=default)
      except ValueError as err:
        return _jsonError(
          str(err), queryParams, status=getattr(err, 'status', 400), encoder=encoder, default=default)
      except Exception as err:
        return _jsonError(
          str(err), queryParams, status=getattr(err, 'status', 500), encoder=encoder, default=default)

    return wrapped_f

  # used like @jsonResponse
  if args:
    return decorator(args[0])

  # used like @jsonResponse(encoder=DjangoJSONEncoder)
  return decorator


class HttpError(Exception):
  def __init__(self, message, status=500):
    super(HttpError, self).__init__(message)
    self.status=status


def _jsonResponse(data, queryParams, status=200, encoder=None, default=None):
  if isinstance(data, HttpResponse):
    return data

  if not queryParams:
    queryParams = {}

  return HttpResponse(
    json.dumps(
      data,
      indent=(2 if queryParams.get('pretty') else None),
      sort_keys=bool(queryParams.get('pretty')),
      cls=encoder,
      default=default
    ),
    content_type='application/json',
    status=status
  )


def _jsonError(message, queryParams, status=500, encoder=None, default=None):
  return _jsonResponse(
    {'error': message}, queryParams, status=status, encoder=encoder, default=default)


def parseHost(host_string):
    s = host_string.strip()
    bidx = s.rfind(']:')    # find closing bracket and following colon.
    cidx = s.find(':')
    if s.startswith('[') and bidx is not None:
        server = s[1:bidx]
        port = s[bidx + 2:]
    elif cidx is not None:
        server = s[:cidx]
        port = s[cidx + 1:]
    else:
        raise ValueError("Invalid host string \"%s\"" % host_string)

    if ':' in port:
        port, _, instance = port.partition(':')
    else:
        instance = None

    return server, int(port), instance


def parseHosts(host_strings):
    return [parseHost(host_string) for host_string in host_strings]



import codecs
import time

from six.moves.urllib.parse import urlencode, urlsplit, parse_qs

from django.conf import settings
from django.core.cache import cache

from graphite.http_pool import http
from graphite.intervals import Interval, IntervalSet
from graphite.logger import log
from graphite.node import LeafNode, BranchNode
from graphite.render.hashing import compactHash
from graphite.util import unpickle, logtime, is_local_interface, json, msgpack, BufferedHTTPReader

from graphite.finders.utils import BaseFinder
from graphite.readers.remote import RemoteReader


class RemoteFinder(BaseFinder):
    local = False

    @classmethod
    def factory(cls):
        finders = []
        for host in settings.CLUSTER_SERVERS:
            if settings.REMOTE_EXCLUDE_LOCAL and is_local_interface(cls.parse_host(host)['host']):
                continue
            finders.append(cls(host))
        return finders

    @staticmethod
    def parse_host(host):
        if host.startswith('http://') or host.startswith('https://'):
            parsed = urlsplit(host)
        else:
            scheme = 'https' if settings.INTRACLUSTER_HTTPS else 'http'
            parsed = urlsplit(scheme + '://' + host)

        return {
          'host': parsed.netloc,
          'url': '%s://%s%s' % (parsed.scheme, parsed.netloc, parsed.path),
          'params': {key: value[-1] for (key, value) in parse_qs(parsed.query).items()},
        }

    def __init__(self, host):
        parsed = self.parse_host(host)
        self.host = parsed['host']
        self.url = parsed['url']
        self.params = parsed['params']
        self.last_failure = 0
        self.tags = not self.params.get('noTags')

    @property
    def disabled(self):
        return time.time() - self.last_failure < settings.REMOTE_RETRY_DELAY

    def fail(self):
        self.last_failure = time.time()

    @logtime
    def find_nodes(self, query, timer=None):
        timer.set_msg(
            'host: {host}, query: {query}'.format(
                host=self.host,
                query=query))

        log.debug("RemoteFinder.find_nodes(host=%s, query=%s) called" % (self.host, query))

        # prevent divide by 0
        cacheTTL = settings.FIND_CACHE_DURATION or 1
        if query.startTime:
            start = query.startTime - (query.startTime % cacheTTL)
        else:
            start = ""

        if query.endTime:
            end = query.endTime - (query.endTime % cacheTTL)
        else:
            end = ""

        cacheKey = "find:%s:%s:%s:%s" % (self.host, compactHash(query.pattern), start, end)

        results = cache.get(cacheKey)
        if results is not None:
            log.debug(
                "RemoteFinder.find_nodes(host=%s, query=%s) using cached result" %
                (self.host, query))
        else:
            url = '/metrics/find/'

            query_params = [
                ('local', self.params.get('local', '1')),
                ('format', self.params.get('format', 'pickle')),
                ('query', query.pattern),
            ]
            if query.startTime:
                query_params.append(('from', int(query.startTime)))

            if query.endTime:
                query_params.append(('until', int(query.endTime)))

            result = self.request(
                url,
                fields=query_params,
                headers=query.headers,
                timeout=settings.FIND_TIMEOUT)

            results = self.deserialize(result)

            cache.set(cacheKey, results, settings.FIND_CACHE_DURATION)

        # We don't use generator here, this function may be run as a job in a thread pool, using a generator has the following risks:
        # 1. Generators are lazy, if we don't iterator the returned generator in the job, the real execution(network operations,
        #    time-consuming) are very likely be triggered in the calling thread, losing the effect of thread pool;
        # 2. As function execution is delayed, the job manager can not catch job runtime exception as expected/designed;
        nodes = []
        for node_info in results:
            # handle both 1.x and 0.9.x output
            path = node_info.get('path') or node_info.get('metric_path')
            is_leaf = node_info.get('is_leaf') or node_info.get('isLeaf')
            intervals = node_info.get('intervals') or []
            if not isinstance(intervals, IntervalSet):
                intervals = IntervalSet(
                    [Interval(interval[0], interval[1]) for interval in intervals])

            node_info = {
                'is_leaf': is_leaf,
                'path': path,
                'intervals': intervals,
            }

            if is_leaf:
                reader = RemoteReader(self, node_info)
                node = LeafNode(path, reader)
            else:
                node = BranchNode(path)

            node.local = False
            nodes.append(node)

        return nodes

    def fetch(self, patterns, start_time, end_time, now=None, requestContext=None):
        reader = RemoteReader(self, {}, bulk_query=patterns)
        return reader.fetch_multi(start_time, end_time, now, requestContext)

    def get_index(self, requestContext):
        url = '/metrics/index.json'

        headers = requestContext.get('forwardHeaders')

        result = self.request(
            url,
            fields=[
              ('local', self.params.get('local', '1')),
            ],
            headers=headers,
            timeout=settings.FIND_TIMEOUT)

        try:
            reader = codecs.getreader('utf-8')
            results = json.load(reader(result))
        except Exception as err:
            self.fail()
            log.exception(
                "RemoteFinder[%s] Error decoding index response from %s: %s" %
                (self.host, result.url_full, err))
            raise Exception("Error decoding index response from %s: %s" % (result.url_full, err))
        finally:
            result.release_conn()

        return results

    def auto_complete_tags(self, exprs, tagPrefix=None, limit=None, requestContext=None):
        """
        Return auto-complete suggestions for tags based on the matches for the specified expressions, optionally filtered by tag prefix
        """
        if limit is None:
            limit = settings.TAGDB_AUTOCOMPLETE_LIMIT

        fields = [
            ('tagPrefix', tagPrefix or ''),
            ('limit', str(limit)),
            ('local', self.params.get('local', '1')),
        ]
        for expr in exprs:
            fields.append(('expr', expr))

        result = self.request(
            '/tags/autoComplete/tags',
            fields,
            headers=requestContext.get('forwardHeaders') if requestContext else None,
            timeout=settings.FIND_TIMEOUT)
        try:
            reader = codecs.getreader('utf-8')
            results = json.load(reader(result))
        except Exception as err:
            self.fail()
            log.exception(
                "RemoteFinder[%s] Error decoding autocomplete tags response from %s: %s" %
                (self.host, result.url_full, err))
            raise Exception("Error decoding autocomplete tags response from %s: %s" % (result.url_full, err))
        finally:
            result.release_conn()

        return results

    def auto_complete_values(self, exprs, tag, valuePrefix=None, limit=None, requestContext=None):
        """
        Return auto-complete suggestions for tags and values based on the matches for the specified expressions, optionally filtered by tag and/or value prefix
        """
        if limit is None:
            limit = settings.TAGDB_AUTOCOMPLETE_LIMIT

        fields = [
            ('tag', tag or ''),
            ('valuePrefix', valuePrefix or ''),
            ('limit', str(limit)),
            ('local', self.params.get('local', '1')),
        ]
        for expr in exprs:
            fields.append(('expr', expr))

        result = self.request(
            '/tags/autoComplete/values',
            fields,
            headers=requestContext.get('forwardHeaders') if requestContext else None,
            timeout=settings.FIND_TIMEOUT)
        try:
            reader = codecs.getreader('utf-8')
            results = json.load(reader(result))
        except Exception as err:
            self.fail()
            log.exception(
                "RemoteFinder[%s] Error decoding autocomplete values response from %s: %s" %
                (self.host, result.url_full, err))
            raise Exception("Error decoding autocomplete values response from %s: %s" % (result.url_full, err))
        finally:
            result.release_conn()

        return results

    def request(self, path, fields=None, headers=None, timeout=None):
        url = "%s%s" % (self.url, path)
        url_full = "%s?%s" % (url, urlencode(fields))

        try:
            result = http.request(
                'POST' if settings.REMOTE_STORE_USE_POST else 'GET',
                url,
                fields=fields,
                headers=headers,
                timeout=timeout,
                preload_content=False)
        except BaseException as err:
            self.fail()
            log.exception("RemoteFinder[%s] Error requesting %s: %s" % (self.host, url_full, err))
            raise Exception("Error requesting %s: %s" % (url_full, err))

        if result.status != 200:
            result.release_conn()
            self.fail()
            log.exception(
                "RemoteFinder[%s] Error response %d from %s" % (self.host, result.status, url_full))
            raise Exception("Error response %d from %s" % (result.status, url_full))

        result.url_full = url_full

        # reset last failure time so that retried fetches can re-enable a remote
        self.last_failure = 0

        log.debug("RemoteFinder[%s] Fetched %s" % (self.host, url_full))
        return result

    def deserialize(self, result):
        """
        Based on configuration, either stream-deserialize a response in settings.REMOTE_BUFFER_SIZE chunks,
        or read the entire payload and use inline deserialization.
        :param result: an http response object
        :return: deserialized response payload from cluster server
        """
        start = time.time()
        try:
            should_buffer = settings.REMOTE_BUFFER_SIZE > 0
            measured_reader = MeasuredReader(BufferedHTTPReader(result, settings.REMOTE_BUFFER_SIZE))

            if should_buffer:
                log.debug("Using streaming deserializer.")
                reader = BufferedHTTPReader(measured_reader, settings.REMOTE_BUFFER_SIZE)
                return self._deserialize_stream(reader, result.getheader('content-type'))

            log.debug("Using inline deserializer for small payload")
            return self._deserialize_buffer(measured_reader.read(), result.getheader('content-type'))
        except Exception as err:
            self.fail()
            log.exception(
                "RemoteFinder[%s] Error decoding response from %s: %s" %
                (self.host, result.url_full, err))
            raise Exception("Error decoding response from %s: %s" % (result.url_full, err))
        finally:
            log.debug("Processed %d bytes in %f seconds." % (measured_reader.bytes_read, time.time() - start))
            result.release_conn()

    @staticmethod
    def _deserialize_buffer(byte_buffer, content_type):
        if content_type == 'application/x-msgpack':
            data = msgpack.unpackb(byte_buffer, encoding='utf-8')
        else:
            data = unpickle.loads(byte_buffer)

        return data

    @staticmethod
    def _deserialize_stream(stream, content_type):
        if content_type == 'application/x-msgpack':
            data = msgpack.load(stream, encoding='utf-8')
        else:
            data = unpickle.load(stream)

        return data


class MeasuredReader(object):
    def __init__(self, reader):
        self.reader = reader
        self.bytes_read = 0

    def read(self, amt=None):
        b = b''
        try:
            if amt:
                b = self.reader.read(amt)
            else:
                b = self.reader.read()
            return b
        finally:
            self.bytes_read += len(b)
